#!/usr/bin/env python
"""
Bridge Example

This example is quite similar to the Hello example
but displays a hello form both the parent and child
processing demonstrating how IPC works using the Bridge.
"""

from os import getpid

from circuits import Component, Event, ipc


class hello(Event):
    """hello Event"""


class Child(Component):
    def hello(self):
        return f'Hello from child with pid {getpid()}'


class App(Component):
    def init(self):
        Child().start(process=True, link=self)

    def ready(self, *args):
        x = yield self.call(hello())
        yield print(x)

        y = yield self.call(ipc(hello()))
        yield print(y)

        raise SystemExit(0)

    def hello(self):
        return f'Hello from parent with pid {getpid()}'


App().run()



from urllib.parse import parse_qsl


class QueryStringToken:
    ARRAY = 'ARRAY'
    OBJECT = 'OBJECT'
    KEY = 'KEY'


class QueryStringParser:
    def __init__(self, data):
        self.result = {}

        sorted_pairs = self._sorted_from_string(data) if isinstance(data, str) else self._sorted_from_obj(data)

        [self.process(x) for x in sorted_pairs]

    def _sorted_from_string(self, data):
        stage1 = parse_qsl(data, keep_blank_values=True)
        stage2 = [(x[0].strip(), x[1].strip()) for x in stage1]
        return sorted(stage2, key=lambda p: p[0])

    def _sorted_from_obj(self, data):
        # data is a list of the type generated by parse_qsl
        if isinstance(data, list):
            items = data
        else:
            # complex objects:
            try:
                # django.http.QueryDict,
                items = [(i[0], j) for i in data.lists() for j in i[1]]
            except AttributeError:
                # webob.multidict.MultiDict
                # werkzeug.datastructures.MultiDict
                items = data.items()

        return sorted(items, key=lambda p: p[0])

    def process(self, pair):
        key = pair[0]
        value = pair[1]

        # faster than invoking a regex
        try:
            key.index('[')
            self.parse(key, value)
            return
        except ValueError:
            pass

        try:
            key.index('.')
            self.parse(key, value)
            return
        except ValueError:
            pass

        self.result[key] = value

    def parse(self, key, value):
        ref = self.result
        tokens = self.tokens(key)

        for token in tokens:
            token_type, key = token

            if token_type == QueryStringToken.ARRAY:
                if key not in ref:
                    ref[key] = []
                ref = ref[key]

            elif token_type == QueryStringToken.OBJECT:
                if key not in ref:
                    ref[key] = {}
                ref = ref[key]

            elif token_type == QueryStringToken.KEY:
                try:
                    ref = ref[key]
                    next(tokens)
                # TypeError is for pet[]=lucy&pet[]=ollie
                # if the array key is empty a type error will be raised
                except (IndexError, KeyError, TypeError):
                    # the index didn't exist
                    # so we look ahead to see what we are setting
                    # there is not a next token
                    # set the value
                    try:
                        next_token = next(tokens)

                        if next_token[0] == QueryStringToken.ARRAY:
                            ref.append([])
                            ref = ref[key]
                        elif next_token[0] == QueryStringToken.OBJECT:
                            try:
                                ref[key] = {}
                            except IndexError:
                                ref.append({})

                            ref = ref[key]
                    except StopIteration:
                        try:
                            ref.append(value)
                        except AttributeError:
                            ref[key] = value
                        return

    def tokens(self, key):
        buf = ''
        for char in key:
            if char == '[':
                yield QueryStringToken.ARRAY, buf
                buf = ''

            elif char == '.':
                yield QueryStringToken.OBJECT, buf
                buf = ''

            elif char == ']':
                try:
                    yield QueryStringToken.KEY, int(buf)
                    buf = ''
                except ValueError:
                    yield QueryStringToken.KEY, None
            else:
                buf = buf + char

        if len(buf) > 0:
            yield QueryStringToken.KEY, buf



"""Internet Relay Chat Utilities"""

from re import compile as compile_regex


PREFIX = compile_regex('([^!].*)!(.*)@(.*)')
COLOR_CODE = compile_regex(r'(?:(\d\d?)(?:(,)(\d\d?))?)?')
COLOR = compile_regex(r'\x03(?:(\d\d?)(?:,(\d\d?))?)?')


class Error(Exception):
    """Error Exception"""


def strip(s, color=False):
    """
    strip(s, color=False) -> str

    Strips the : from the start of a string
    and optionally also strips all colors if
    color is True.

    :param s str: string to process
    :param color bool: whether to strip colors

    :returns str: returns processes string
    """
    if len(s) > 0 and s[0] == ':':
        s = s[1:]
    if color:
        s = s.replace('\x01', '')
        s = s.replace('\x02', '')  # bold
        s = s.replace('\x1d', '')  # italics
        s = s.replace('\x1f', '')  # underline
        s = s.replace('\x1e', '')  # strikethrough
        s = s.replace('\x11', '')  # monospace
        s = s.replace('\x16', '')  # reverse color
        s = COLOR.sub('', s)  # color codes
        s = s.replace('\x03', '')  # color
        s = s.replace('\x0f', '')  # reset
    return s


def joinprefix(nick, user, host):
    """
    Join the parts of a prefix

    :param nick str: nickname
    :param user str: username
    :param host str: hostname

    :returns str: a string in the form of <nick>!<user>@<host>
    """
    return '{}!{}@{}'.format(nick or '', user or '', host or '')


def parseprefix(prefix):
    """
    Parse a prefix into it's parts

    :param prefix str: prefix to parse

    :returns tuple: tuple of strings in the form of (nick, user, host)
    """
    m = PREFIX.match(prefix)

    if m is not None:
        return m.groups()
    return prefix or None, None, None


def parsemsg(s, encoding='utf-8'):
    """
    Parse an IRC Message from s

    :param s bytes: bytes to parse
    :param encoding str: encoding to use (Default: utf-8)

    :returns tuple: parsed message in the form of (prefix, command, args)
    """
    s = s.decode(encoding, 'replace')

    prefix = ''
    trailing = []

    if s and s[0] == ':':
        prefix, s = s[1:].split(' ', 1)

    prefix = parseprefix(prefix)

    if s.find(' :') != -1:
        s, trailing = s.split(' :', 1)
        args = s.split()
        args.append(trailing)
    else:
        args = s.split()

    args = iter(args)
    command = next(args, None)
    command = command and str(command)

    return prefix, command, list(args)


def irc_color_to_ansi(data, reset=True):
    """Maps IRC color codes to ANSI terminal escape sequences"""

    def ansi(*seq):
        return '\33[{}m'.format(';'.join(f'{x:02}' for x in seq if x))

    ansi_default_fg = 39
    ansi_default_bg = 49
    color_map_fg = {
        0: 37,
        1: 30,
        2: 34,
        3: 32,
        4: 31,
        5: 36,
        6: 35,
        7: 33,
        8: 93,
        9: 92,
        10: 36,
        11: 96,
        12: 94,
        13: 95,
        14: 90,
        15: 37,
        16: 52,
        17: 94,
        18: 100,
        19: 58,
        20: 22,
        21: 29,
        22: 23,
        23: 24,
        24: 17,
        25: 54,
        26: 53,
        27: 89,
        28: 88,
        29: 130,
        30: 142,
        31: 64,
        32: 28,
        33: 35,
        34: 30,
        35: 25,
        36: 18,
        37: 91,
        38: 90,
        39: 125,
        40: 124,
        41: 166,
        42: 184,
        43: 106,
        44: 34,
        45: 49,
        46: 37,
        47: 33,
        48: 19,
        49: 129,
        50: 127,
        51: 161,
        52: 196,
        53: 208,
        54: 226,
        55: 154,
        56: 46,
        57: 86,
        58: 51,
        59: 75,
        60: 21,
        61: 171,
        62: 201,
        63: 198,
        64: 203,
        65: 215,
        66: 227,
        67: 191,
        68: 83,
        69: 122,
        70: 87,
        71: 111,
        72: 63,
        73: 177,
        74: 207,
        75: 205,
        76: 217,
        77: 223,
        78: 229,
        79: 193,
        80: 157,
        81: 158,
        82: 159,
        83: 153,
        84: 147,
        85: 183,
        86: 219,
        87: 212,
        88: 16,
        89: 233,
        90: 235,
        91: 237,
        92: 239,
        93: 241,
        94: 244,
        95: 247,
        96: 250,
        97: 254,
        98: 231,
        99: ansi_default_fg,
    }
    color_map_bg = {
        0: 47,
        1: 40,
        2: 44,
        3: 42,
        4: 41,
        5: 46,
        6: 45,
        7: 43,
        8: 103,
        10: 46,
        14: 97,
        15: 47,
        99: ansi_default_bg,
    }

    enable_char = {
        '\x02': 1,
        '\x1d': 3,
        '\x1e': 9,
        '\x1f': 4,
        '\x16': 7,
    }
    revert_char = {
        '\x02': 22,
        '\x1d': 23,
        '\x1f': 24,
        '\x16': 27,
        '\x1e': 29,
    }

    def escape(data):
        ignore = set()
        start = []
        current_fg = ansi_default_fg
        current_bg = ansi_default_bg
        for i, char in enumerate(data):
            if i in ignore:
                continue
            if char == '\x0f':  # reset
                start = []
                yield ansi(0)
            elif char in start and char in revert_char:
                start.remove(char)
                yield ansi(revert_char[char])
            elif char in enable_char:
                start.append(char)
                yield ansi(enable_char[char])
            elif char == '\x03':
                i += 1
                m = COLOR_CODE.match(data[i : i + 5])
                colors = []
                if m:
                    fg, has_bg, bg = m.groups()
                    if fg:
                        ignore.update(range(i, i + len(fg)))
                        colors.append(color_map_fg.get(int(fg), current_fg))
                        current_fg = int(fg)
                    if has_bg:
                        ignore.update(range(i + len(fg), i + len(fg) + 1))
                    if bg:
                        ignore.update(range(i + len(fg) + 1, i + len(fg) + 1 + len(bg)))
                        colors.append(color_map_bg.get(int(bg), current_bg))
                        current_bg = int(bg)
                if char in start:
                    start.remove(char)
                if colors:
                    start.append(char)
                    yield ansi(*colors)
                else:
                    yield ansi(ansi_default_fg, ansi_default_bg)
            # elif char == u'\x04':
            #     if char[i + 1:i + 6].isdigit():
            #         ignore.update(range(i, i + 6))
            #     # TODO: parse hex representation
            # elif char == u'\x11':  # monospace
            #     start.append(char)
            else:
                yield char
        if start and reset:
            yield ansi(0)

    return ''.join(escape(data))



from io import BytesIO

from circuits.core import BaseComponent, Event, handler


class request(Event):
    """request Event"""


class response(Event):
    """response Event"""


class ResponseObject:
    def __init__(self, headers, status, version):
        self.headers = headers
        self.status = status
        self.version = version

        self.body = BytesIO()

        # TODO: This sucks :/ Avoiding the circuit import here :/
        from circuits.web.constants import HTTP_STATUS_CODES

        self.reason = HTTP_STATUS_CODES[self.status]

    def __repr__(self):
        return '<Response {:d} {} {} ({:d})>'.format(
            self.status,
            self.reason,
            self.headers.get('Content-Type'),
            len(self.body.getvalue()),
        )

    def read(self):
        return self.body.read()


class HTTP(BaseComponent):
    channel = 'web'

    def __init__(self, encoding='utf-8', channel=channel):
        super().__init__(channel=channel)

        self._encoding = encoding

        # TODO: This sucks :/ Avoiding the circuit import here :/
        from circuits.web.parsers import HttpParser

        self._parser = HttpParser(1, True)

    @handler('read')
    def _on_client_read(self, data):
        self._parser.execute(data, len(data))
        if (
            self._parser.is_message_complete()
            or self._parser.is_upgrade()
            or (self._parser.is_headers_complete() and self._parser._clen == 0)
        ):
            status = self._parser.get_status_code()
            version = self._parser.get_version()
            headers = self._parser.get_headers()

            res = ResponseObject(headers, status, version)
            res.body.write(self._parser.recv_body())
            res.body.seek(0)
            self.fire(response(res))

            # TODO: This sucks :/ Avoiding the circuit import here :/
            from circuits.web.parsers import HttpParser

            self._parser = HttpParser(1, True)



#!/usr/bin/python -i
from circuits import Component, Event, Manager


class foo(Event):
    """foo Event"""

    channels = ('a',)


class bar(Event):
    """bar Event"""


class A(Component):
    channel = 'a'

    def foo(self):
        return 'Foo'


class B(Component):
    channel = 'b'

    def foo(self):
        return 'Hello World!'


class C(Component):
    channel = 'c'

    def foo(self):
        return self.fire(bar())

    def bar(self):
        return 'Bar'


def test():
    m = Manager() + A() + B() + C()

    while len(m):
        m.flush()

    # Rely on Event.channels
    x = m.fire(foo())
    m.flush()
    assert x.value == 'Foo'

    # Explicitly specify the channel
    x = m.fire(foo(), 'b')
    m.flush()
    assert x.value == 'Hello World!'

    # Explicitly specify a set of channels
    x = m.fire(foo(), 'a', 'b')
    m.flush()
    assert x.value == ['Foo', 'Hello World!']

    # Rely on self.channel
    x = m.fire(foo(), 'c')
    m.flush()
    m.flush()
    assert x.value == 'Bar'



#!/usr/bin/env python
from setuptools import setup


def parse_requirements(filename):
    with open(filename) as f:
        for line in f:
            if line and line[0] != '#':
                yield line.strip()


setup(
    name='wiki',
    version='dev',
    description='circuits wiki demo',
    scripts=('wiki.py',),
    install_requires=list(parse_requirements('requirements.txt')),
    zip_safe=True,
)



"""
Manager Repr Tests

Test Manager's representation string.
"""

import os
from threading import current_thread
from time import sleep

import pytest

from circuits import Component, Manager


class App(Component):
    def test(self, event, *args, **kwargs):
        pass


def test_main():
    id = f'{os.getpid()}:{current_thread().name}'

    m = Manager()
    assert repr(m) == '<Manager/ %s (queued=0) [S]>' % id

    app = App()
    app.register(m)
    s = repr(m)
    assert s == '<Manager/ %s (queued=1) [S]>' % id

    m.start()

    pytest.wait_for(m, '_running', True)
    sleep(0.1)

    s = repr(m)
    assert s == '<Manager/ %s (queued=0) [R]>' % id

    m.stop()

    pytest.wait_for(m, '_Manager__thread', None)

    s = repr(m)
    assert s == '<Manager/ %s (queued=0) [S]>' % id



"""
Workers

Worker is a component used to perform "work" in independent threads or
processes. Simply create an instance of Worker() with either `process=True`
to create a pool of workers using sub-processes for CPU-bound work or `False`
(the default) for a thread pool of workers for I/O bound work.

Then fire `task()` events with a function and `*args` and `**kwargs` to pass
to the function when called from within the workers.
"""

from multiprocessing import Pool as ProcessPool, cpu_count
from multiprocessing.pool import ThreadPool
from threading import current_thread
from weakref import WeakKeyDictionary

from .components import BaseComponent
from .events import Event
from .handlers import handler


DEFAULT_WORKERS = 10


class task(Event):
    """
    task Event

    This Event is used to initiate a new task to be performed by a Worker

    :param f: The function to be executed.
    :type  f: function

    :param args: Arguments to pass to the function
    :type  args: tuple

    :param kwargs: Keyword Arguments to pass to the function
    :type  kwargs: dict
    """

    success = True
    failure = True

    def __init__(self, f, *args, **kwargs):
        """x.__init__(...) initializes x; see x.__class__.__doc__ for signature"""
        super().__init__(f, *args, **kwargs)


class Worker(BaseComponent):
    """
    A thread/process Worker Component

    This Component creates a pool of workers (either a thread or process)
    and executures the supplied function from a `task()` event passing
    supplied arguments and keyword-arguments to the function.

    A `task_success` event is fired upon successful execution of the function
    and `task_failure` if it failed and threw an exception. The `task()` event
    can also be "waited" upon by using the `.call()` and `.wait()` primitives.

    :param process: True to start this Worker as a process (Thread otherwise)
    :type process: bool
    """

    channel = 'worker'

    def init(self, process=False, workers=None, channel=channel):
        if not hasattr(current_thread(), '_children'):
            current_thread()._children = WeakKeyDictionary()

        self.workers = workers or (cpu_count() if process else DEFAULT_WORKERS)
        Pool = ProcessPool if process else ThreadPool
        self.pool = Pool(self.workers)

    @handler('stopped', 'unregistered', channel='*')
    def _on_stopped(self, event, *args):
        if event.name == 'unregistered' and args[0] is not self:
            return

        self.pool.close()
        self.pool.join()

    @handler('task')
    def _on_task(self, f, *args, **kwargs):
        result = self.pool.apply_async(f, args, kwargs)
        while not result.ready():
            yield
        yield result.get()



"""Core

This package contains the essential core parts of the circuits framework.
"""

from .bridge import Bridge, ipc
from .components import BaseComponent, Component
from .debugger import Debugger
from .events import Event
from .handlers import handler, reprhandler
from .loader import Loader
from .manager import Manager, TimeoutError, sleep
from .timers import Timer
from .values import Value
from .workers import Worker, task


__all__ = (
    'handler',
    'BaseComponent',
    'Component',
    'Event',
    'task',
    'Worker',
    'ipc',
    'Bridge',
    'Debugger',
    'Timer',
    'Manager',
    'TimeoutError',
)

# flake8: noqa
# pylama: skip=1



#!/usr/bin/env python
from http.client import HTTPConnection
from os import path

from circuits.web import Controller

from .conftest import DOCROOT
from .helpers import HTTPError, quote, urlopen


class Root(Controller):
    def index(self):
        return 'Hello World!'


def test(webapp):
    f = urlopen(webapp.server.http.base)
    s = f.read()
    assert s == b'Hello World!'


def test_404(webapp):
    try:
        urlopen('%s/foo' % webapp.server.http.base)
    except HTTPError as e:
        assert e.code == 404
        assert e.msg == 'Not Found'
    else:
        assert False


def test_file(webapp):
    url = '%s/static/helloworld.txt' % webapp.server.http.base
    f = urlopen(url)
    s = f.read().strip()
    assert s == b'Hello World!'


def test_largefile(webapp):
    url = '%s/static/largefile.txt' % webapp.server.http.base
    f = urlopen(url)
    s = f.read().strip()
    assert s == open(path.join(DOCROOT, 'largefile.txt'), 'rb').read()


def test_file404(webapp):
    try:
        urlopen('%s/static/foo.txt' % webapp.server.http.base)
    except HTTPError as e:
        assert e.code == 404
        assert e.msg == 'Not Found'
    else:
        assert False


def test_directory(webapp):
    f = urlopen('%s/static/' % webapp.server.http.base)
    s = f.read()
    assert b'helloworld.txt' in s


def test_file_quoting(webapp):
    url = '{:s}{:s}'.format(webapp.server.http.base, quote('/static/#foobar.txt'))
    f = urlopen(url)
    s = f.read().strip()
    assert s == b'Hello World!'


def test_range(webapp):
    connection = HTTPConnection(webapp.server.host, webapp.server.port)

    connection.request('GET', '%s/static/largefile.txt' % webapp.server.http.base, headers={'Range': 'bytes=0-100'})
    response = connection.getresponse()
    assert response.status == 206
    s = response.read()
    assert s == open(path.join(DOCROOT, 'largefile.txt'), 'rb').read(101)


def test_ranges(webapp):
    connection = HTTPConnection(webapp.server.host, webapp.server.port)

    connection.request('GET', '%s/static/largefile.txt' % webapp.server.http.base, headers={'Range': 'bytes=0-50,51-100'})
    response = connection.getresponse()
    assert response.status == 206

    # TODO: Complete this test.
    # ``response.read()`` is a multipart/bytes-range
    # See: Issue #59
    # s = response.read()
    # assert s == open(path.join(DOCROOT, "largefile.txt"), "rb").read(101)


def test_unsatisfiable_range1(webapp):
    connection = HTTPConnection(webapp.server.host, webapp.server.port)

    connection.request(
        'GET', '%s/static/largefile.txt' % webapp.server.http.base, headers={'Range': 'bytes=0-100,100-10000,0-1'}
    )
    response = connection.getresponse()
    assert response.status == 416


# TODO: Implement this test and condition
# e.g: For a 10 byte file; Range: bytes=0-1,2-3,4-5,6-7,8-9
# def test_unsatisfiable_range2(webapp):
#    connection = HTTPConnection(webapp.server.host, webapp.server.port)
#
#    connection.request("GET", "%s/static/largefile.txt" % (webapp.server.http.base,
#       headers={"Range": "bytes=0-100,100-10000,0-1"}))
#    response = connection.getresponse()
#    assert response.status == 416



#!/usr/bin/env python
import re
import sys
from io import StringIO
from socket import gaierror, gethostbyname, gethostname

from circuits.web import Controller, Logger

from .helpers import urlopen


class DummyLogger:
    def __init__(self):
        super().__init__()

        self.message = None

    def info(self, message):
        self.message = message


class Root(Controller):
    def index(self):
        return 'Hello World!'


def test_file(webapp):
    logfile = StringIO()
    logger = Logger(file=logfile)
    logger.register(webapp)

    f = urlopen(webapp.server.http.base)
    s = f.read()
    assert s == b'Hello World!'

    logfile.seek(0)
    s = logfile.read().strip()

    try:
        address = gethostbyname(gethostname())
    except gaierror:
        address = '127.0.0.1'

    d = {}
    d['h'] = address
    d['l'] = '-'
    d['u'] = '-'
    d['r'] = 'GET / HTTP/1.1'
    d['s'] = '200'
    d['b'] = '12'
    d['f'] = ''
    d['a'] = 'Python-urllib/%s' % sys.version[:3]

    keys = list(d.keys())

    for k in keys:
        if d[k] and d[k].startswith('127.'):
            # loopback network: 127.0.0.0/8
            assert re.search(r'127(\.[0-9]{1,3}){3}', s)
        else:
            assert d[k] in s

    logfile.close()
    logger.unregister()


def test_logger(webapp):
    logobj = DummyLogger()
    logger = Logger(logger=logobj)
    logger.register(webapp)

    f = urlopen(webapp.server.http.base)
    s = f.read()
    assert s == b'Hello World!'

    s = logobj.message

    try:
        address = gethostbyname(gethostname())
    except gaierror:
        address = '127.0.0.1'

    d = {}
    d['h'] = address
    d['l'] = '-'
    d['u'] = '-'
    d['r'] = 'GET / HTTP/1.1'
    d['s'] = '200'
    d['b'] = '12'
    d['f'] = ''
    d['a'] = 'Python-urllib/%s' % sys.version[:3]

    keys = list(d.keys())

    for k in keys:
        if d[k] and d[k].startswith('127.'):
            # loopback network: 127.0.0.0/8
            assert re.search(r'127(\.[0-9]{1,3}){3}', s)
        else:
            assert d[k] in s

    logger.unregister()


def test_filename(webapp, tmpdir):
    logfile = str(tmpdir.ensure('logfile'))
    logger = Logger(file=logfile)
    logger.register(webapp)

    logfile = open(logfile)

    f = urlopen(webapp.server.http.base)
    s = f.read()
    assert s == b'Hello World!'

    logfile.seek(0)
    s = logfile.read().strip()

    try:
        address = gethostbyname(gethostname())
    except gaierror:
        address = '127.0.0.1'

    d = {}
    d['h'] = address
    d['l'] = '-'
    d['u'] = '-'
    d['r'] = 'GET / HTTP/1.1'
    d['s'] = '200'
    d['b'] = '12'
    d['f'] = ''
    d['a'] = 'Python-urllib/%s' % sys.version[:3]

    keys = list(d.keys())

    for k in keys:
        if d[k] and d[k].startswith('127.'):
            # loopback network: 127.0.0.0/8
            assert re.search(r'127(\.[0-9]{1,3}){3}', s)
        else:
            assert d[k] in s

    logfile.close()
    logger.unregister()



#!/usr/bin/env python
from http.client import HTTPConnection

from circuits.web import Controller


class Root(Controller):
    def index(self):
        return 'Hello World!'


def test_bad_header(webapp):
    connection = HTTPConnection(webapp.server.host, webapp.server.port)
    connection.connect()

    connection.putrequest('GET', '/', 'HTTP/1.1')
    connection.putheader('Connection', 'close')
    connection._output(b'X-Foo')  # Bad Header
    connection.endheaders()

    response = connection.getresponse()
    assert response.status == 400
    assert response.reason == 'Bad Request'

    connection.close()



"""
Daemon Component

Component to daemonize a system into the background and detach it from its
controlling PTY. Supports PID file writing, logging stdin, stdout and stderr
and changing the current working directory.
"""

from os import _exit, chdir, closerange, dup2, fork, getpid, remove, setsid, umask
from os.path import isabs
from resource import RLIM_INFINITY, RLIMIT_NOFILE, getrlimit
from sys import stderr, stdin, stdout

from circuits.core import Component, Event, handler


class daemonize(Event):
    """daemonize Event"""


class daemonized(Event):
    """daemonized Event"""


class deletepid(Event):
    """deletepid Event"""


class writepid(Event):
    """writepid Event"""


class Daemon(Component):
    """
    Daemon Component

    :param pidfile: .pid filename
    :type  pidfile: str or unicode

    :param stdin:   filename to log stdin
    :type  stdin:   str or unicode

    :param stdout:  filename to log stdout
    :type  stdout:  str or unicode

    :param stderr:  filename to log stderr
    :type  stderr:  str or unicode
    """

    channel = 'daemon'

    def init(self, pidfile, path='/', stdin=None, stdout=None, stderr=None, channel=channel):
        assert isabs(path), 'path must be absolute'

        self.pidfile = pidfile
        self.path = path

        self.stdin = stdin if stdin is not None and isabs(stdin) else '/dev/null'

        self.stdout = stdout if stdout is not None and isabs(stdout) else '/dev/null'

        self.stderr = stderr if stderr is not None and isabs(stderr) else '/dev/null'

    def deletepid(self):
        remove(self.pidfile)

    def writepid(self):
        with open(self.pidfile, 'w') as fd:
            fd.write(str(getpid()))

    def daemonize(self):
        try:
            pid = fork()
            if pid > 0:
                # exit first parent
                _exit(0)
        except OSError as e:
            stderr.write(f'fork #1 failed: {e.errno:d} ({e})\n')

            raise SystemExit(1)

        # decouple from parent environment
        chdir(self.path)
        setsid()
        umask(0)

        # do second fork
        try:
            pid = fork()
            if pid > 0:
                # exit from second parent
                _exit(0)
        except OSError as e:
            stderr.write(f'fork #2 failed: {e.errno:d} ({e})\n')

            raise SystemExit(1)

        # redirect standard file descriptors
        stdout.flush()
        stderr.flush()

        maxfd = getrlimit(RLIMIT_NOFILE)[1]
        if maxfd == RLIM_INFINITY:
            maxfd = 2048

        closerange(0, maxfd)

        si = open(self.stdin)
        so = open(self.stdout, 'a+')
        se = open(self.stderr, 'a+')

        dup2(si.fileno(), stdin.fileno())
        dup2(so.fileno(), stdout.fileno())
        dup2(se.fileno(), stderr.fileno())

        self.fire(writepid())
        self.fire(daemonized(self))

    def registered(self, component, manager):
        if component == self and manager.root.running:
            self.fire(daemonize())

    @handler('started', priority=100.0, channel='*')
    def on_started(self, component):
        if component is not self:
            self.fire(daemonize())



"""
Internet Relay Chat Protocol

This package implements the Internet Relay Chat Protocol
or commonly known as IRC. Support for both server and client
is implemented.
"""

from .commands import *  # noqa
from .events import reply, response  # noqa
from .message import Message  # noqa
from .numerics import *  # noqa
from .protocol import IRC  # noqa
from .utils import (  # noqa
    irc_color_to_ansi,
    joinprefix,
    parsemsg,
    parseprefix,
    strip,
)


sourceJoin = joinprefix
sourceSplit = parseprefix

# pylama:skip=1



#!/usr/bin/env python
import os

import pytest

from circuits import Component, handler


try:
    from circuits.io.notify import Notify
except ImportError:
    pytest.importorskip('pyinotify')


class App(Component):
    def init(self, *args, **kwargs):
        self.created_status = False

    @handler('created', channel='notify')
    def created(self, *args, **kwargs):
        self.created_status = True


class Creator:
    def __init__(self, app, watcher, tmpdir, timeout=0.5):
        self.app = app
        self.watcher = watcher
        self.tmpdir = tmpdir
        self.timeout = timeout

    def create(self, *targets, **kwargs):
        assert_created = kwargs.get('assert_created', True)
        target = os.path.join(*targets)
        self.tmpdir.ensure(target, dir=kwargs.get('dir', False))
        self.watcher.wait('created', timeout=self.timeout)
        assert self.app.created_status == assert_created
        # Reset for next call
        self.watcher.clear()
        self.app.created_status = False


@pytest.fixture()
def app(manager, watcher):
    app = App().register(manager)
    yield app
    # Unregister application on test end
    app.unregister()
    watcher.wait('unregistered')


@pytest.fixture()
def notify(app, watcher):
    notify = Notify().register(app)
    watcher.wait('registered')
    return notify


@pytest.fixture()
def creator(app, watcher, tmpdir):
    return Creator(app, watcher, tmpdir)


# TESTS


def test_notify_file(notify, tmpdir, creator):
    # Add a path to the watch
    notify.add_path(str(tmpdir))

    # Test creation and detection of a file
    creator.create('helloworld.txt')

    # Remove the path from the watch
    notify.remove_path(str(tmpdir))

    # Test creation and NON detection of a file
    creator.create('helloworld2.txt', assert_created=False)


def test_notify_dir(notify, tmpdir, creator):
    # Add a path to the watch
    notify.add_path(str(tmpdir))

    # Test creation and detection of a file
    creator.create('hellodir', dir=True)

    # Remove the path from the watch
    notify.remove_path(str(tmpdir))

    # Test creation and NON detection of a file
    creator.create('hellodir2', dir=True, assert_created=False)


def test_notify_subdir_recursive(notify, tmpdir, creator):
    # Add a subdir
    subdir = 'sub'
    tmpdir.ensure(subdir, dir=True)

    # Add a path to the watch
    notify.add_path(str(tmpdir), recursive=True)

    # Test creation and detection of a file in subdir
    creator.create(subdir, 'helloworld.txt', assert_created=True)


@pytest.mark.xfail(reason='pyinotify issue #133')
def test_notify_subdir_recursive_remove_path(notify, tmpdir, creator):
    # This is logically the second part of the above test,
    # but pyinotify fails on rm_watch(...., rec=True)

    # Add a subdir
    subdir = 'sub'
    tmpdir.ensure(subdir, dir=True)

    # Add a path to the watch
    notify.add_path(str(tmpdir), recursive=True)

    # Remove the path from the watch
    notify.remove_path(str(tmpdir), recursive=True)

    # Test creation and NON detection of a file in subdir
    creator.create(subdir, 'helloworld2.txt', assert_created=False)


def test_notify_subdir_recursive_auto_add(notify, tmpdir, creator):
    # Add a path to the watch
    notify.add_path(str(tmpdir), recursive=True)

    # Create/detect subdirectory
    subdir = 'sub'
    creator.create(subdir, dir=True, assert_created=True)

    # Create/detect file in subdirectory
    creator.create(subdir, 'helloworld.txt', assert_created=True)

    # Skip notify.remove_path() because pyinotify is broken


def test_notify_subdir_recursive_no_auto_add(notify, tmpdir, creator):
    # Add a path to the watch
    notify.add_path(str(tmpdir), recursive=True, auto_add=False)

    # Create/detect subdirectory
    subdir = 'sub'
    creator.create(subdir, dir=True, assert_created=True)

    # Create, not detect file in subdirectory
    creator.create(subdir, 'helloworld.txt', assert_created=False)

    # Skip notify.remove_path() because pyinotify is broken



#!/usr/bin/env python

from time import sleep

import pytest

from circuits import Component, Event
from circuits.net.events import close
from circuits.net.sockets import UDPServer
from circuits.node import Node


pytestmark = pytest.mark.skipif(pytest.PLATFORM == 'win32', reason='Broken on Windows')


class return_value(Event):
    success = True


class App(Component):
    def return_value(self, event):
        print('Hello client!', event.channels)


@pytest.fixture()
def bind(manager, watcher):
    server = UDPServer(0).register(manager)
    assert watcher.wait('ready', channel='server')

    host, port = server.host, server.port

    server.fire(close())
    assert watcher.wait('closed', channel='server')

    server.unregister()
    assert watcher.wait('unregistered', channel='server')

    return host, port


@pytest.fixture()
def app(manager, watcher, bind):
    server = Node(port=bind[1], server_ip=bind[0])
    server.register(manager)
    server.bind = bind
    assert watcher.wait('registered', channel='node')

    return server


def test_auto_reconnect(app, watcher, manager):
    # add client
    client = App().register(manager)
    node = Node().register(client)
    chan = node.add('client1', *app.bind, reconnect_delay=1, connect_timeout=1)
    assert watcher.wait('connected', channel=chan)
    watcher.clear()

    # close server
    app.fire(close(), app.channel)
    assert watcher.wait('closed', channel=app.channel)
    watcher.clear()

    # client gets an unreachable
    assert watcher.wait('connect', channel=chan)
    assert watcher.wait('unreachable', channel=chan)
    watcher.clear()

    # start a new server
    node2 = Node(port=app.bind[1], server_ip=app.bind[0])
    node2.register(manager)
    assert watcher.wait('ready', channel=node2.channel)
    watcher.clear()

    assert watcher.wait('connected', channel=chan)

    client.unregister()


def test_server_send_all(app, watcher, manager):
    client1 = App().register(manager)
    node1 = Node().register(client1)
    chan = node1.add('client1', *app.bind)
    assert watcher.wait('connected', channel=chan)

    client2 = App().register(manager)
    node2 = Node().register(client2)
    chan = node2.add('client2', *app.bind)
    assert watcher.wait('connected', channel=chan)

    event = return_value()
    app.server.send_all(event)
    assert watcher.wait('return_value')

    client1.unregister()
    client2.unregister()


def test_server_send(app, watcher, manager):
    client1 = App().register(manager)
    node1 = Node().register(client1)
    chan1 = node1.add('client1', *app.bind)
    assert watcher.wait('connected', channel=chan1)

    client2 = App().register(manager)
    node2 = Node().register(client2)
    chan2 = node2.add('client2', *app.bind)
    assert watcher.wait('connected', channel=chan2)

    event = return_value()
    app.server.send(event, app.server.get_socks()[0], no_result=True)
    assert watcher.wait('return_value')

    client1.unregister()
    client2.unregister()


def test_server_send_multicast(app, watcher, manager):
    client1 = App().register(manager)
    node1 = Node().register(client1)
    chan1 = node1.add('client1', *app.bind)
    assert watcher.wait('connected', channel=chan1)
    watcher.clear()

    client2 = App().register(manager)
    node2 = Node().register(client2)
    chan2 = node2.add('client2', *app.bind)
    assert watcher.wait('connected', channel=chan2)
    watcher.clear()

    client3 = App().register(manager)
    node3 = Node().register(client3)
    chan3 = node3.add('client3', *app.bind)
    assert watcher.wait('connected', channel=chan3)
    watcher.clear()

    event = return_value()
    app.server.send_to(event, app.server.get_socks())
    assert watcher.wait('return_value')

    for _ in range(3):
        if watcher.count('return_value') == 3:
            break
        sleep(1)
    assert watcher.count('return_value') == 3

    client1.unregister()
    client2.unregister()
    client3.unregister()



"""
Logger Component

This module implements Logger Components.
"""

import datetime
import os
import sys
from io import IOBase

from circuits.core import BaseComponent, handler


def formattime():
    return datetime.datetime.now().strftime('[%d/%b/%Y:%H:%M:%S]')


class Logger(BaseComponent):
    channel = 'web'

    format = '%(h)s %(l)s %(u)s %(t)s "%(r)s" %(s)s %(b)s "%(f)s" "%(a)s"'

    def __init__(self, file=None, logger=None, **kwargs):
        super().__init__(**kwargs)

        if isinstance(file, str):
            self.file = open(os.path.abspath(os.path.expanduser(file)), 'a')
        elif isinstance(file, IOBase) or hasattr(file, 'write'):
            self.file = file
        else:
            self.file = sys.stdout

        self.logger = logger

    @handler('response_success')
    def log_response(self, response_event, value):
        response = response_event.args[0]
        self.log(response)

    def log(self, response):
        request = response.request
        remote = request.remote
        outheaders = response.headers
        inheaders = request.headers or {}

        protocol = 'HTTP/%d.%d' % request.protocol

        host = inheaders.get('X-Forwarded-For', (remote.name or remote.ip))

        atoms = {
            'h': host,
            'l': '-',
            'u': getattr(request, 'login', None) or '-',
            't': formattime(),
            'r': f'{request.method} {request.path} {protocol}',
            's': int(response.status),
            'b': outheaders.get('Content-Length', '') or '-',
            'f': inheaders.get('Referer', ''),
            'a': inheaders.get('User-Agent', ''),
        }
        for k, v in list(atoms.items()):
            if isinstance(v, str):
                v = v.encode('utf8')
            elif not isinstance(v, str):
                v = str(v)
            # Fortunately, repr(str) escapes unprintable chars, \n, \t, etc
            # and backslash for us. All we have to do is strip the quotes.
            v = repr(v)[1:-1]
            # Escape double-quote.
            atoms[k] = v.replace('"', '\\"')

        if self.logger is not None:
            self.logger.info(self.format % atoms)
        else:
            self.file.write(self.format % atoms)
            self.file.write('\n')
            self.file.flush()



#!/usr/bin/env python
import sys
from errno import ESRCH
from os import kill
from signal import SIGTERM
from subprocess import Popen
from time import sleep

import pytest

from . import app


pytestmark = pytest.mark.skipif(pytest.PLATFORM == 'win32', reason='Unsupported Platform')


def is_running(pid):
    try:
        kill(pid, 0)
    except OSError as error:
        if error.errno == ESRCH:
            return False
    return True


def wait(pid, timeout=3):
    count = timeout
    while is_running(pid) and count:
        sleep(1)


def test(tmpdir):
    tmpdir.ensure('app.pid')
    pid_path = tmpdir.join('app.pid')

    args = [sys.executable, app.__file__, str(pid_path)]
    Popen(args, env={'PYTHONPATH': ':'.join(sys.path)}).wait()

    sleep(1)

    assert pid_path.check(exists=True, file=True)

    pid = None
    with pid_path.open() as f:
        pid = int(f.read().strip())

    assert isinstance(pid, int)
    assert pid > 0

    kill(pid, SIGTERM)
    wait(pid)



"""Docker Tasks"""

from fabric.api import local, task

from .utils import msg, requires, tobool


TAG = 'prologic/circuits'


@task(default=True)
@requires('docker')
def build(**options):
    """
    Build Docker Image

    Options can be provided to customize the build.
    The following options are supported:

    - rebuild -> Whether to rebuild without a cache.
    - version -> Specific version to tag the image with (Default: latest)
    """
    rebuild = tobool(options.get('rebuild', False))
    version = options.get('version', 'latest')

    tag = f'{TAG:s}:{version:s}'
    args = ['docker', 'build', '-t', tag, '.']

    if rebuild:
        args.insert(-1, '--no-cache')

    with msg('Building Image'):
        local(' '.join(args))


@task()
@requires('docker')
def publish():
    """Publish Docker Image"""
    args = ['docker', 'push', TAG]

    with msg('Pushing Image'):
        local(' '.join(args))


@task()
@requires('docker')
def run():
    """Run Docker Container"""
    args = ['docker', 'run', '-i', '-t', '--rm', TAG]

    local(' '.join(args))



#!/usr/bin/env python
from json import loads

from circuits.web import JSONController, Sessions

from .helpers import CookieJar, HTTPCookieProcessor, build_opener, urlopen


class Root(JSONController):
    def index(self):
        return {'success': True, 'message': 'Hello World!'}

    def test_sessions(self, name=None):
        if name:
            with self.session as data:
                data['name'] = name
        else:
            name = self.session.get('name', 'World!')

        return {'success': True, 'message': 'Hello %s' % name}


def test(webapp):
    f = urlopen(webapp.server.http.base)
    data = f.read()
    data = data.decode('utf-8')
    d = loads(data)
    assert d['success']
    assert d['message'] == 'Hello World!'


def test_sessions(webapp):
    Sessions().register(webapp)

    cj = CookieJar()
    opener = build_opener(HTTPCookieProcessor(cj))

    f = opener.open('%s/test_sessions' % webapp.server.http.base)
    data = f.read()
    data = data.decode('utf-8')
    d = loads(data)
    assert d['success']
    assert d['message'] == 'Hello World!'

    f = opener.open('%s/test_sessions/test' % webapp.server.http.base)
    data = f.read()
    data = data.decode('utf-8')
    d = loads(data)
    assert d['success']
    assert d['message'] == 'Hello test'

    f = opener.open('%s/test_sessions' % webapp.server.http.base)
    data = f.read()
    data = data.decode('utf-8')
    d = loads(data)
    assert d['success']
    assert d['message'] == 'Hello test'



#!/usr/bin/env python
from circuits import Component, Event
from circuits.web import Controller

from .helpers import urlopen


class foo(Event):
    """foo Event"""


class App(Component):
    channel = 'app'

    def foo(self):
        return 'Hello World!'


class Root(Controller):
    def index(self):
        value = yield self.call(foo(), 'app')
        yield value.value


def test(webapp):
    app = App().register(webapp)
    try:
        f = urlopen(webapp.server.http.base)
        s = f.read()
        assert s == b'Hello World!'
    finally:
        app.unregister()



"""
Utils macros

Utility macros
"""

from inspect import getdoc


def macros(macro, environ, *args, **kwargs):
    """Return a list of available macros"""
    macros = environ['macros'].items()
    s = '\n'.join([f'== {k} ==\n{getdoc(v)}\n' for k, v in macros])

    return environ['parser'].generate(s, environ=environ)



#!/usr/bin/env python
from http.client import HTTPConnection

from circuits.web import Controller


class Root(Controller):
    def index(self):
        return 'Hello World!'


def test_GET(webapp):
    connection = HTTPConnection(webapp.server.host, webapp.server.port)

    connection.request('GET', '/')
    response = connection.getresponse()
    assert response.status == 200
    assert response.reason == 'OK'
    s = response.read()
    assert s == b'Hello World!'


def test_HEAD(webapp):
    connection = HTTPConnection(webapp.server.host, webapp.server.port)

    connection.request('HEAD', '/')
    response = connection.getresponse()
    assert response.status == 200
    assert response.reason == 'OK'
    s = response.read()
    assert s == b''



#!/usr/bin/env python
"""circuits Hello World"""

from circuits import Component, Event


class hello(Event):
    """hello Event"""


class terminate(Event):
    """terminate Event"""


class App(Component):
    def hello(self):
        """Hello Event Handler"""
        print('Hello World!')

    def started(self, *args):
        """
        Started Event Handler

        This is fired internally when your application starts up
        and can be used to trigger events that only occur once
        during startup.
        """
        self.fire(hello())  # Fire hello Event
        self.fire(terminate())

    def terminate(self):
        raise SystemExit(0)  # Terminate the Application


App().run()



#!/usr/bin/env python
import sys
from os.path import abspath

from circuits import Component
from circuits.app import Daemon


try:
    from coverage import coverage

    HAS_COVERAGE = True
except ImportError:
    HAS_COVERAGE = False


class App(Component):
    def init(self, pidfile):
        self.pidfile = pidfile

    def started(self, *args):
        Daemon(self.pidfile).register(self)

    def prepare_unregister(self, *args):
        return


def main():
    if HAS_COVERAGE:
        _coverage = coverage(data_suffix=True)
        _coverage.start()

    args = iter(sys.argv)
    next(args)  # executable

    pidfile = next(args)  # pidfile

    pidfile = abspath(pidfile)

    App(pidfile).run()

    if HAS_COVERAGE:
        _coverage.stop()
        _coverage.save()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
Clone of the standard UNIX "ping" command.

This example shows how you can utilize some of the buitlin I/O components
in circuits to write a very simple clone of the standard UNIX "ping" command.
This example merely wraps the standard UNIX "/usr/bin/ping" command in a subprocess
using the ``circuits.io.Process`` Component for Asyncchronous I/O communications with
the process.
"""

import sys

from circuits import Component, Debugger
from circuits.io import Process, stdout, write


class Ping(Component):
    # This adds the already instantiated stdout instnace
    stdout = stdout

    def init(self, host):
        self.p = Process(['/bin/ping', host]).register(self)
        self.p.start()

    def read(self, data):
        """
        read Event Handler

        This is fired by the File Component when there is data to be read
        from the underlying file that was opened.
        """
        self.fire(write(data), stdout)


# Start and "run" the system.
app = Ping(sys.argv[1])
Debugger().register(app)
app.run()



from circuits import Event


class connected_to(Event):
    """Connected to a peer"""

    def __init__(self, connection_name, hostname, port, client_channel, client_obj):
        """
        :param connection_name:    Connection name.
        :type connection_name:     str

        :param hostname:    hostname of the remote system.
        :type hostname:     str

        :param port:    connection port of the remote system.
        :type port:     int

        :param client_channel:    Channel used for client event.
        :type client_channel:     str

        :param client_obj:    Client object.
        :type client_obj:     :class:`circuits.net.sockets.Client`
        """
        super().__init__(connection_name, hostname, port, client_channel, client_obj)


class disconnected_from(connected_to):
    """Disconnected from a peer"""


class remote(Event):
    """send event to a peer"""

    def __init__(self, event, connection_name, channel=None):
        """
        :param event:    Event to execute remotely.
        :type event:     :class:`circuits.core.events.Event`

        :param connection_name:    Connection name.
        :type connection_name:     str

        :param channel:    Remote channel (channel to use on peer).
        :type channel:     str
        """
        super().__init__(event, connection_name, channel=channel)



"""
Component Repr Tests

Test Component's representation string.
"""

import os
from threading import current_thread

from circuits import Component, Event


class App(Component):
    def test(self, event, *args, **kwargs):
        pass


class test(Event):
    pass


def test_main():
    id = f'{os.getpid()}:{current_thread().name}'
    app = App()

    assert repr(app) == '<App/* %s (queued=0) [S]>' % id

    app.fire(test())
    assert repr(app) == '<App/* %s (queued=1) [S]>' % id

    app.flush()
    assert repr(app) == '<App/* %s (queued=0) [S]>' % id


def test_non_str_channel():
    id = f'{os.getpid()}:{current_thread().name}'
    app = App(channel=(1, 1))

    assert repr(app) == '<App/(1, 1) %s (queued=0) [S]>' % id

    app.fire(test())
    assert repr(app) == '<App/(1, 1) %s (queued=1) [S]>' % id

    app.flush()
    assert repr(app) == '<App/(1, 1) %s (queued=0) [S]>' % id



#!/usr/bin/env python
"""
IRC Bot Example

This example shows how to use several components in circuits as well
as one of the builtin networking protocols. This IRC Bot simply connects
to the Libera.Chat IRC Network and joins the #circuits channel. It will also
echo anything privately messages to it in response.
"""

import sys

from circuits import Component, Debugger
from circuits.net.sockets import TCPClient, connect
from circuits.protocols.irc import ERR_NICKNAMEINUSE, ERR_NOMOTD, IRC, JOIN, NICK, PRIVMSG, RPL_ENDOFMOTD, USER


class Bot(Component):
    # Define a separate channel so we can create many instances of ``Bot``
    channel = 'ircbot'

    def init(self, host='irc.libera.chat', port='6667', channel=channel):
        self.host = host
        self.port = int(port)

        # Add TCPClient and IRC to the system.
        TCPClient(channel=self.channel).register(self)
        IRC(channel=self.channel).register(self)

    def ready(self, component):
        """
        Ready Event

        This event is triggered by the underlying ``TCPClient`` Component
        when it is ready to start making a new connection.
        """
        self.fire(connect(self.host, self.port))

    def connected(self, host, port):
        """
        connected Event

        This event is triggered by the underlying ``TCPClient`` Component
        when a successfully connection has been made.
        """
        self.fire(NICK('circuits'))
        self.fire(USER('circuits', 'circuits', host, 'Test circuits IRC Bot'))

    def disconnected(self):
        """
        disconnected Event

        This event is triggered by the underlying ``TCPClient`` Component
        when the connection is lost.
        """
        raise SystemExit(0)

    def numeric(self, source, numeric, *args):
        """
        Numeric Event

        This event is triggered by the ``IRC`` Protocol Component when we have
        received an IRC Numberic Event from server we are connected to.
        """
        if numeric == ERR_NICKNAMEINUSE:
            self.fire(NICK(f'{args[0]:s}_'))
        elif numeric in (RPL_ENDOFMOTD, ERR_NOMOTD):
            self.fire(JOIN('#circuits'))

    def privmsg(self, source, target, message):
        """
        Message Event

        This event is triggered by the ``IRC`` Protocol Component for each
        message we receieve from the server.
        """
        if target.startswith('#'):
            self.fire(PRIVMSG(target, message))
        else:
            self.fire(PRIVMSG(source[0], message))


# Configure and run the system
bot = Bot(*sys.argv[1:])

Debugger().register(bot)

# To register a 2nd ``Bot`` instance. Simply use a separate channel.
# Bot(*sys.argv[1:], channel="foo").register(bot)

bot.run()



"""
Global Constants

This module implements required shared global constants.
"""

from circuits import __version__


SERVER_PROTOCOL = (1, 1)
SERVER_VERSION = 'circuits.web/%s' % __version__
SERVER_URL = 'http://circuitsweb.com/'

DEFAULT_ERROR_MESSAGE = """\
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html>
 <head>
  <meta http-equiv="Content-Type" content="text/html; charset=utf-8"></meta>
  <title>%(code)d %(name)s</title>
  <style type="text/css">
   #powered_by {
    margin-top: 20px;
    border-top: 2px solid black;
    font-style: italic;
   }

   #traceback {
    color: red;
   }
  </style>
 </head>
 <body>
  <h1>%(name)s</h1>
  %(description)s
  <pre id="traceback">%(traceback)s</pre>
  %(powered_by)s
 </body>
</html>
"""

POWERED_BY = """
<div id="powered_by">
    <span>Powered by <a href="%(url)s">%(version)s</a></span>
</div>
"""

HTTP_STATUS_CODES = {
    100: 'Continue',
    101: 'Switching Protocols',
    102: 'Processing',
    200: 'OK',
    201: 'Created',
    202: 'Accepted',
    203: 'Non Authoritative Information',
    204: 'No Content',
    205: 'Reset Content',
    206: 'Partial Content',
    207: 'Multi Status',
    226: 'IM Used',
    300: 'Multiple Choices',
    301: 'Moved Permanently',
    302: 'Found',
    303: 'See Other',
    304: 'Not Modified',
    305: 'Use Proxy',
    307: 'Temporary Redirect',
    400: 'Bad Request',
    401: 'Unauthorized',
    402: 'Payment Required',
    403: 'Forbidden',
    404: 'Not Found',
    405: 'Method Not Allowed',
    406: 'Not Acceptable',
    407: 'Proxy Authentication Required',
    408: 'Request Timeout',
    409: 'Conflict',
    410: 'Gone',
    411: 'Length Required',
    412: 'Precondition Failed',
    413: 'Request Entity Too Large',
    414: 'Request URI Too Long',
    415: 'Unsupported Media Type',
    416: 'Requested Range Not Satisfiable',
    417: 'Expectation Failed',
    418: 'I"m a teapot',
    422: 'Unprocessable Entity',
    423: 'Locked',
    424: 'Failed Dependency',
    426: 'Upgrade Required',
    449: 'Retry With',
    500: 'Internal Server Error',
    501: 'Not Implemented',
    502: 'Bad Gateway',
    503: 'Service Unavailable',
    504: 'Gateway Timeout',
    505: 'HTTP Version Not Supported',
    507: 'Insufficient Storage',
    510: 'Not Extended',
}



#!/usr/bin/env python
from datetime import datetime
from optparse import OptionParser

from circuits import Component, Debugger, Event, Timer
from circuits.node import Node


__version__ = '0.0.1'

USAGE = '%prog [options]'
VERSION = '%prog v' + __version__


def parse_options():
    parser = OptionParser(usage=USAGE, version=VERSION)

    parser.add_option(
        '-b',
        '--bind',
        action='store',
        type='string',
        default='0.0.0.0:8000',
        dest='bind',
        help='Bind to address:[port]',
    )

    parser.add_option(
        '-d',
        '--debug',
        action='store_true',
        default=False,
        dest='debug',
        help='Enable debug mode',
    )

    opts, args = parser.parse_args()

    return opts, args


class send_all_event(Event):
    def __init__(self, infos):
        super().__init__(infos)


class NodeServer(Component):
    def init(self, args, opts):
        if opts.debug:
            Debugger().register(self)

        if ':' in opts.bind:
            address, port = opts.bind.split(':')
            port = int(port)
        else:
            address, port = opts.bind, 8000

        self.node = Node(port=port, server_ip=address).register(self)

    def connect(self, sock, host, port):
        print('Peer connected: %s:%d' % (host, port))

    def disconnect(self, sock):
        print('Peer disconnected: %s' % sock)

    def ready(self, server, bind):
        print('Server ready: %s:%d' % bind)
        Timer(3, Event.create('send_all'), persist=True).register(self)

    def send_all(self):
        event = send_all_event(str(datetime.now()))
        print('send: %s' % event)
        self.node.server.send_all(event)


def main():
    opts, args = parse_options()

    # Configure and "run" the System.
    NodeServer(args, opts).run()


if __name__ == '__main__':
    main()



"""Circuits Library - Web

circuits.web contains the circuits full stack web server that is HTTP
and WSGI compliant.
"""

from .controllers import BaseController, Controller, expose
from .dispatchers import XMLRPC, Dispatcher, Static, VirtualHosts
from .errors import forbidden, httperror, notfound, redirect
from .events import request, response, stream
from .loggers import Logger
from .servers import BaseServer, Server
from .sessions import Sessions
from .url import URL, parse_url


try:
    from .dispatchers import JSONRPC
except ImportError:
    pass

try:
    from .controllers import JSONController
except ImportError:
    pass

# flake8: noqa
# pylama: skip=1



"""py.test config"""

import sys
import threading
from collections import deque
from collections.abc import Callable
from time import sleep

import pytest

from circuits import BaseComponent, Debugger, Manager, handler
from circuits.core.manager import TIMEOUT


class Watcher(BaseComponent):
    def init(self):
        self._lock = threading.Lock()
        self.events = deque()

    @handler(channel='*', priority=999.9)
    def _on_event(self, event, *args, **kwargs):
        with self._lock:
            self.events.append(event)

    def clear(self):
        self.events.clear()

    def wait(self, name, channel=None, timeout=30.0):
        for _i in range(int(timeout / TIMEOUT)):
            with self._lock:
                for event in self.events:
                    if (
                        event.name == name
                        and event.waitingHandlers == 0
                        and ((channel is None) or (channel in event.channels))
                    ):
                        return True
            sleep(TIMEOUT)
        return False

    def count(self, name, channel=None, n=1, timeout=30.0):
        n = 0
        with self._lock:
            for event in self.events:
                if event.name == name and event.waitingHandlers == 0 and ((channel is None) or (channel in event.channels)):
                    n += 1
        return n


class Flag:
    status = False


def call_event_from_name(manager, event, event_name, *channels):
    fired = False
    value = None
    for _r in manager.waitEvent(event_name):
        if not fired:
            fired = True
            value = manager.fire(event, *channels)
        sleep(0.1)
    return value


def call_event(manager, event, *channels):
    return call_event_from_name(manager, event, event.name, *channels)


class WaitEvent:
    def __init__(self, manager, name, channel=None, timeout=30.0):
        if channel is None:
            channel = getattr(manager, 'channel', None)

        self.timeout = timeout
        self.manager = manager

        flag = Flag()

        @handler(name, channel=channel)
        def on_event(self, *args, **kwargs):
            flag.status = True

        self.handler = self.manager.addHandler(on_event)
        self.flag = flag

    def wait(self):
        try:
            for _i in range(int(self.timeout / TIMEOUT)):
                if self.flag.status:
                    return True
                sleep(TIMEOUT)
        finally:
            self.manager.removeHandler(self.handler)


def wait_for(obj, attr, value=True, timeout=30.0):
    from circuits.core.manager import TIMEOUT

    for _i in range(int(timeout / TIMEOUT)):
        if isinstance(value, Callable):
            if value(obj, attr):
                return True
        elif getattr(obj, attr) == value:
            return True
        sleep(TIMEOUT)
    return None


class SimpleManager(Manager):
    def tick(self, timeout=-1):
        self._running = False
        return super().tick(timeout)


@pytest.fixture()
def simple_manager(request):
    manager = SimpleManager()
    Debugger(events=request.config.option.verbose).register(manager)
    return manager


@pytest.fixture()
def manager(request):
    manager = Manager()

    def finalizer():
        manager.stop()

    request.addfinalizer(finalizer)

    waiter = WaitEvent(manager, 'started')
    manager.start()
    assert waiter.wait()

    Debugger(events=request.config.option.verbose).register(manager)

    return manager


@pytest.fixture()
def watcher(request, manager):
    watcher = Watcher().register(manager)

    def finalizer():
        waiter = WaitEvent(manager, 'unregistered')
        watcher.unregister()
        waiter.wait()

    request.addfinalizer(finalizer)

    return watcher


for key, value in {
    'WaitEvent': WaitEvent,
    'wait_for': wait_for,
    'call_event': call_event,
    'PLATFORM': sys.platform,
    'PYVER': sys.version_info[:3],
    'call_event_from_name': call_event_from_name,
}.items():
    setattr(pytest, key, value)



#!/usr/bin/env python
"""
Example IRC Client

A basic IRC client with a very basic console interface.

For usage type:

   ./ircclient.py --help

"""

import os
from optparse import OptionParser
from socket import gethostname

from circuits import Component, Debugger, __version__ as systemVersion, handler
from circuits.io import stdin
from circuits.net.events import connect
from circuits.net.sockets import TCPClient
from circuits.protocols.irc import IRC, JOIN, NICK, PRIVMSG, USER


USAGE = '%prog [options] host [port]'
VERSION = '%prog v' + systemVersion


def parse_options():
    parser = OptionParser(usage=USAGE, version=VERSION)

    parser.add_option(
        '-n',
        '--nick',
        action='store',
        default=os.environ['USER'],
        dest='nick',
        help='Nickname to use',
    )

    parser.add_option(
        '-d',
        '--debug',
        action='store_true',
        default=False,
        dest='debug',
        help='Enable debug verbose logging',
    )

    parser.add_option(
        '-c',
        '--channel',
        action='store',
        default='#circuits',
        dest='channel',
        help='Channel to join',
    )

    opts, args = parser.parse_args()

    if len(args) < 1:
        parser.print_help()
        raise SystemExit(1)

    return opts, args


class Client(Component):
    # Set a separate channel in case we want multiple ``Client`` instances.
    channel = 'ircclient'

    def init(self, host, port=6667, opts=None):
        self.host = host
        self.port = port
        self.opts = opts
        self.hostname = gethostname()

        self.nick = opts.nick
        self.ircchannel = opts.channel

        # Add TCPClient and IRC to the system.
        TCPClient(channel=self.channel).register(self)
        IRC(channel=self.channel).register(self)

        # Enable Debugging?
        if opts.debug:
            Debugger().register(self)

    def ready(self, component):
        """
        ready Event

        This event is triggered by the underlying ``TCPClient`` Component
        when it is ready to start making a new connection.
        """
        self.fire(connect(self.host, self.port))

    def connected(self, host, port):
        """
        connected Event

        This event is triggered by the underlying ``TCPClient`` Component
        when a successfully connection has been made.
        """
        print('Connected to %s:%d' % (host, port))

        nick = self.nick
        hostname = self.hostname
        name = f'{nick} on {hostname} using circuits/{systemVersion}'

        self.fire(NICK(nick))
        self.fire(USER(nick, nick, self.hostname, name))

    def disconnected(self):
        """
        disconnected Event

        This event is triggered by the underlying ``TCPClient`` Component
        when the connection has been disconnected.
        """
        print('Disconnecetd from %s:%d' % (self.host, self.port))

        raise SystemExit(0)

    def numeric(self, source, numeric, *args):
        """
        numeric Event

        This event is triggered by the ``IRC`` Protocol Component when we have
        received an IRC Numberic Event from server we are connected to.
        """
        if numeric == 1:
            self.fire(JOIN(self.ircchannel))
        elif numeric == 433:
            self.nick = newnick = '%s_' % self.nick
            self.fire(NICK(newnick))

    def join(self, source, channel):
        """
        join Event

        This event is triggered by the ``IRC`` Protocol Component when a
        user has joined a channel.
        """
        if source[0].lower() == self.nick.lower():
            print('Joined %s' % channel)
        else:
            print(
                '--> %s (%s) has joined %s'
                % (
                    source[0],
                    '@'.join(source[1:]),
                    channel,
                ),
            )

    def notice(self, source, target, message):
        """
        notice Event

        This event is triggered by the ``IRC`` Protocol Component for each
        notice we receieve from the server.
        """
        print(f'-{source[0]}- {message}')

    def privmsg(self, source, target, message):
        """
        privmsg Event

        This event is triggered by the ``IRC`` Protocol Component for each
        message we receieve from the server.
        """
        if target[0] == '#':
            print(f'<{source[0]}> {message}')
        else:
            print(f'-{source[0]}- {message}')

    @handler('read', channel='stdin')
    def stdin_read(self, data):
        """
        read Event (on channel ``stdin``)

        This is the event handler for ``read`` events specifically from the
        ``stdin`` channel. This is triggered each time stdin has data that
        it has read.
        """
        data = data.strip().decode('utf-8')

        print(f'<{self.nick:s}> {data:s}')
        self.fire(PRIVMSG(self.ircchannel, data))


def main():
    opts, args = parse_options()

    host = args[0]
    port = int(args[1]) if len(args) > 1 else 6667

    # Configure and run the system.
    client = Client(host, port, opts=opts)
    stdin.register(client)
    client.run()


if __name__ == '__main__':
    main()



"""
Process

This module implements a wrapper for basic ``subprocess.Popen`` functionality.
"""

from io import BytesIO
from subprocess import PIPE, Popen

from circuits import BaseComponent, Event, handler
from circuits.core.manager import TIMEOUT

from .events import close, started, write
from .file import File


class terminated(Event):
    """
    terminated Event

    This Event is sent when a process is completed

    :param args:  (process)
    :type  tuple: tuple
    """

    def __init__(self, *args):
        super().__init__(*args)


class Process(BaseComponent):
    channel = 'process'

    def init(self, args, cwd=None, shell=False):
        self.args = args
        self.cwd = cwd
        self.shell = shell

        self.p = None
        self.stderr = BytesIO()
        self.stdout = BytesIO()

        self._status = None
        self._terminated = False

        self._stdout_closed = False
        self._stderr_closed = False

        self._stdin = None
        self._stderr = None
        self._stdout = None

        self._stdin_closed_handler = None
        self._stderr_read_handler = None
        self._stdout_read_handler = None
        self._stderr_closed_handler = None
        self._stdout_closed_handler = None

    def start(self):
        self.p = Popen(
            self.args,
            cwd=self.cwd,
            shell=self.shell,
            stdin=PIPE,
            stderr=PIPE,
            stdout=PIPE,
        )

        self.stderr = BytesIO()
        self.stdout = BytesIO()

        self._status = None

        self._stdin = File(
            self.p.stdin,
            channel=f'{self.p.pid:d}.stdin',
        ).register(self)

        self._stderr = File(
            self.p.stderr,
            channel=f'{self.p.pid:d}.stderr',
        ).register(self)

        self._stdout = File(
            self.p.stdout,
            channel=f'{self.p.pid:d}.stdout',
        ).register(self)

        self._stderr_read_handler = self.addHandler(
            handler('read', channel=self._stderr.channel)(
                lambda self, data: self.stderr.write(data),
            ),
        )

        self._stdout_read_handler = self.addHandler(
            handler('read', channel=self._stdout.channel)(
                lambda self, data: self.stdout.write(data),
            ),
        )

        self._stderr_closed_handler = self.addHandler(
            handler('closed', channel=self._stderr.channel)(
                lambda self: setattr(self, '_stderr_closed', True),
            ),
        )

        self._stdout_closed_handler = self.addHandler(
            handler('closed', channel=self._stdout.channel)(
                lambda self: setattr(self, '_stdout_closed', True),
            ),
        )

        self.fire(started(self))

    def stop(self):
        if self.p is not None:
            self.p.terminate()

    def kill(self):
        self.p.kill()

    def signal(self, signal):
        self.p.send_signal(signal)

    def wait(self):
        return self.p.wait()

    def write(self, data):
        self.fire(write(data), f'{self.p.pid:d}.stdin')

    @property
    def status(self):
        if getattr(self, 'p', None) is not None:
            return self.p.poll()
        return None

    @handler('generate_events')
    def _on_generate_events(self, event):
        if self.p is not None and self._status is None:
            self._status = self.p.poll()

        if self._status is not None and self._stderr_closed and self._stdout_closed and not self._terminated:
            self._terminated = True
            self.removeHandler(self._stderr_read_handler)
            self.removeHandler(self._stdout_read_handler)
            self.removeHandler(self._stderr_closed_handler)
            self.removeHandler(self._stdout_closed_handler)

            self.fire(terminated(self))
            self.fire(close(), self._stdin.channel, self._stdout.channel, self._stderr.channel)

            event.reduce_time_left(0)
            event.stop()
        else:
            event.reduce_time_left(TIMEOUT)



# a port of xmlrpclib to json....
#
#
# The JSON-RPC client interface is based on the XML-RPC client
#
# Copyright (c) 1999-2002 by Secret Labs AB
# Copyright (c) 1999-2002 by Fredrik Lundh
# Copyright (c) 2006 by Matt Harrison
#
# By obtaining, using, and/or copying this software and/or its
# associated documentation, you agree that you have read, understood,
# and will comply with the following terms and conditions:
#
# Permission to use, copy, modify, and distribute this software and
# its associated documentation for any purpose and without fee is
# hereby granted, provided that the above copyright notice appears in
# all copies, and that both that copyright notice and this permission
# notice appear in supporting documentation, and that the name of
# Secret Labs AB or the author not be used in advertising or publicity
# pertaining to distribution of the software without specific, written
# prior permission.
#
# SECRET LABS AB AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD
# TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANT-
# ABILITY AND FITNESS.  IN NO EVENT SHALL SECRET LABS AB OR THE AUTHOR
# BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY
# DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
# WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
# OF THIS SOFTWARE.
# --------------------------------------------------------------------
import base64
import json
from http.client import HTTPConnection, HTTPSConnection
from urllib.parse import splithost, splittype, splituser, unquote


__version__ = '0.0.1'

ID = 1


def _gen_id():
    global ID
    ID = ID + 1
    return ID


# --------------------------------------------------------------------
# Exceptions

##
# Base class for all kinds of client-side errors.


class Error(Exception):
    """Base class for client errors."""

    def __str__(self):
        return repr(self)


##
# Indicates an HTTP-level protocol error.  This is raised by the HTTP
# transport layer, if the server returns an error code other than 200
# (OK).
#
# @param url The target URL.
# @param errcode The HTTP error code.
# @param errmsg The HTTP error message.
# @param headers The HTTP header dictionary.


class ProtocolError(Error):
    """Indicates an HTTP protocol error."""

    def __init__(self, url, errcode, errmsg, headers, response):
        Error.__init__(self)
        self.url = url
        self.errcode = errcode
        self.errmsg = errmsg
        self.headers = headers
        self.response = response

    def __repr__(self):
        return '<ProtocolError for %s: %s %s>' % (self.url, self.errcode, self.errmsg)


def getparser(encoding):
    un = Unmarshaller(encoding)
    par = Parser(un)
    return par, un


def dumps(params, methodname=None, methodresponse=None, encoding=None, allow_none=0):
    if methodname:
        request = {}
        request['method'] = methodname
        request['params'] = params
        request['id'] = _gen_id()
        return json.dumps(request)
    return None


class Unmarshaller:
    def __init__(self, encoding):
        self.data = None
        self.encoding = encoding

    def feed(self, data):
        if self.data is None:
            self.data = data
        else:
            self.data = self.data + data

    def close(self):
        # try to convert string to json
        return json.loads(self.data.decode(self.encoding))


class Parser:
    def __init__(self, unmarshaller):
        self._target = unmarshaller
        self.data = None

    def feed(self, data):
        if self.data is None:
            self.data = data
        else:
            self.data = self.data + data

    def close(self):
        self._target.feed(self.data)


class _Method:
    # some magic to bind an JSON-RPC method to an RPC server.
    # supports "nested" methods (e.g. examples.getStateName)

    def __init__(self, send, name):
        self.__send = send
        self.__name = name

    def __getattr__(self, name):
        return _Method(self.__send, f'{self.__name}.{name}')

    def __call__(self, *args):
        return self.__send(self.__name, args)


##
# Standard transport class for JSON-RPC over HTTP.
# <p>
# You can create custom transports by subclassing this method, and
# overriding selected methods.


class Transport:
    """Handles an HTTP transaction to an JSON-RPC server."""

    # client identifier (may be overridden)
    user_agent = 'jsonlib.py/%s (by matt harrison)' % __version__

    ##
    # Send a complete request, and parse the response.
    #
    # @param host Target host.
    # @param handler Target PRC handler.
    # @param request_body JSON-RPC request body.
    # @param verbose Debugging flag.
    # @return Parsed response.

    def request(self, host, handler, request_body, encoding, verbose=0):
        # issue JSON-RPC request

        h = self.make_connection(host)
        if verbose:
            h.set_debuglevel(1)

        self.send_request(h, handler, request_body)
        self.send_user_agent(h)
        self.send_content(h, request_body)

        try:
            errcode, errmsg, headers = h.getreply()
            r = h.getfile()
        except AttributeError:
            r = h.getresponse()
            errcode = r.status
            errmsg = r.reason
            headers = r.getheaders()

        if errcode != 200:
            response = r.read()
            raise ProtocolError(
                host + handler,
                errcode,
                errmsg,
                headers,
                response,
            )

        self.verbose = verbose

        try:
            sock = h._conn.sock
        except AttributeError:
            sock = None

        return self._parse_response(r, sock, encoding)

    ##
    # Create parser.
    #
    # @return A 2-tuple containing a parser and a unmarshaller.

    def getparser(self, encoding):
        # get parser and unmarshaller
        return getparser(encoding)

    ##
    # Get authorization info from host parameter
    # Host may be a string, or a (host, x509-dict) tuple; if a string,
    # it is checked for a "user:pw@host" format, and a "Basic
    # Authentication" header is added if appropriate.
    #
    # @param host Host descriptor (URL or (URL, x509 info) tuple).
    # @return A 3-tuple containing (actual host, extra headers,
    #     x509 info).  The header and x509 fields may be None.

    def get_host_info(self, host):
        x509 = {}
        if isinstance(host, tuple):
            host, x509 = host

        auth, host = splituser(host)

        if auth:
            auth = base64.encodestring(unquote(auth))
            auth = ''.join(auth.split())  # get rid of whitespace
            extra_headers = [
                ('Authorization', 'Basic ' + auth),
            ]
        else:
            extra_headers = None

        return host, extra_headers, x509

    ##
    # Connect to server.
    #
    # @param host Target host.
    # @return A connection handle.

    def make_connection(self, host):
        # create a HTTP connection object from a host descriptor
        host, _extra_headers, _x509 = self.get_host_info(host)
        return HTTPConnection(host)

    ##
    # Send request header.
    #
    # @param connection Connection handle.
    # @param handler Target RPC handler.
    # @param request_body JSON-RPC body.

    def send_request(self, connection, handler, request_body):
        connection.putrequest('POST', handler)

    ##
    # Send host name.
    #
    # @param connection Connection handle.
    # @param host Host name.

    def send_host(self, connection, host):
        host, extra_headers, _x509 = self.get_host_info(host)
        connection.putheader('Host', host)
        if extra_headers:
            if isinstance(extra_headers, dict):
                extra_headers = list(extra_headers.items())
            for key, value in extra_headers:
                connection.putheader(key, value)

    ##
    # Send user-agent identifier.
    #
    # @param connection Connection handle.

    def send_user_agent(self, connection):
        connection.putheader('User-Agent', self.user_agent)

    ##
    # Send request body.
    #
    # @param connection Connection handle.
    # @param request_body JSON-RPC request body.

    def send_content(self, connection, request_body):
        connection.putheader('Content-Type', 'text/xml')
        connection.putheader('Content-Length', str(len(request_body)))
        connection.endheaders()
        if request_body:
            connection.send(request_body)

    ##
    # Parse response.
    #
    # @param file Stream.
    # @return Response tuple and target method.

    def parse_response(self, file):
        # compatibility interface
        return self._parse_response(file, None)

    ##
    # Parse response (alternate interface).  This is similar to the
    # parse_response method, but also provides direct access to the
    # underlying socket object (where available).
    #
    # @param file Stream.
    # @param sock Socket handle (or None, if the socket object
    #    could not be accessed).
    # @return Response tuple and target method.

    def _parse_response(self, file, sock, encoding):
        # read response from input file/socket, and parse it

        p, u = self.getparser(encoding)

        while 1:
            response = sock.recv(1024) if sock else file.read(1024)
            if not response:
                break
            if self.verbose:
                print('body:', repr(response))
            p.feed(response)

        file.close()
        p.close()

        return u.close()


##
# Standard transport class for JSON-RPC over HTTPS.


class SafeTransport(Transport):
    """Handles an HTTPS transaction to an JSON-RPC server."""

    # FIXME: mostly untested

    def make_connection(self, host):
        # create a HTTPS connection object from a host descriptor
        # host may be a string, or a (host, x509-dict) tuple
        host, _extra_headers, x509 = self.get_host_info(host)
        try:
            HTTPS = HTTPSConnection
        except AttributeError:
            raise NotImplementedError(
                "your version of httplib doesn't support HTTPS",
            )
        else:
            return HTTPS(host, None, **(x509 or {}))


class ServerProxy:
    def __init__(self, uri, transport=None, encoding=None, verbose=None, allow_none=0):
        utype, uri = splittype(uri)
        if utype not in ('http', 'https'):
            raise OSError('Unsupported JSONRPC protocol')
        self.__host, self.__handler = splithost(uri)
        if not self.__handler:
            self.__handler = '/RPC2'

        if transport is None:
            transport = SafeTransport() if utype == 'https' else Transport()
        self.__transport = transport

        self.__encoding = encoding
        self.__verbose = verbose
        self.__allow_none = allow_none

    def __request(self, methodname, params):
        """Call a method on the remote server"""
        request = dumps(params, methodname, encoding=self.__encoding, allow_none=self.__allow_none)

        response = self.__transport.request(
            self.__host,
            self.__handler,
            request.encode(self.__encoding),
            self.__encoding,
            verbose=self.__verbose,
        )

        if len(response) == 1:
            response = response[0]

        return response

    def __repr__(self):
        return '<JSONProxy for %s%s>' % (self.__host, self.__handler)

    __str__ = __repr__

    def __getattr__(self, name):
        # dispatch
        return _Method(self.__request, name)

    # note: to call a remote object with an non-standard name, use
    # result getattr(server, "strange-python-name")(args)


if __name__ == '__main__':
    s = ServerProxy('http://localhost:8080/foo/', verbose=1)
    c = s.echo('foo bar')
    print(c)
    d = s.bad('other')
    print(d)
    e = s.echo('foo bar', 'baz')
    print(e)
    f = s.echo(5)
    print(f)



#!/usr/bin/env python
# stdlib
import ssl

from circuits import Debugger
from circuits.web import Controller, Server


class Root(Controller):
    def GET(self, peer_cert=None):
        return "Here's your cert %s" % peer_cert


app = Server(
    ('0.0.0.0', 8443),
    secure=True,
    certfile='server-cert.pem',
    keyfile='server-key.pem',
    ca_certs='ca-chain.pem',
    cert_reqs=ssl.CERT_OPTIONAL,
)
Root().register(app)
app += Debugger()
app.run()



"""Dispatchers

This package contains various circuits.web dispatchers
By default a ``circuits.web.Server`` Component uses the
``dispatcher.Dispatcher``
"""

from ..websockets.dispatcher import WebSocketsDispatcher
from .dispatcher import Dispatcher
from .jsonrpc import JSONRPC
from .static import Static
from .virtualhosts import VirtualHosts
from .xmlrpc import XMLRPC


# flake8: noqa
# pylama: skip=1



"""
Static

This modStatic implements a Static dispatcher used to serve up static
resources and an optional apache-style directory listing.
"""

import os
from html import escape
from string import Template
from urllib.parse import quote, unquote

from circuits import BaseComponent, handler
from circuits.web.tools import serve_file


DEFAULT_DIRECTORY_INDEX_TEMPLATE = """
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
 "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en" lang="en">
    <head>
        <meta http-equiv="Content-type" content="text/html; charset=utf-8" />
        <meta http-equiv="Content-Language" content="en-us" />
        <meta name="robots" content="NONE,NOARCHIVE" />
        <title>Index of $directory</title>
    </head>
    <body>
        <h1>Index of $directory</h1>
        <ul>
            $url_up
            $listing
        </ul>
    </body>
</html>
"""

_dirlisting_template = Template(DEFAULT_DIRECTORY_INDEX_TEMPLATE)


class Static(BaseComponent):
    channel = 'web'

    def __init__(self, path=None, docroot=None, defaults=('index.html', 'index.xhtml'), dirlisting=False, **kwargs):
        super().__init__(**kwargs)

        self.path = path
        self.docroot = os.path.abspath(docroot) if docroot is not None else os.path.abspath(os.getcwd())
        self.defaults = defaults
        self.dirlisting = dirlisting

    @handler('request', priority=0.9)
    def _on_request(self, event, request, response):
        if self.path is not None and not request.path.startswith(self.path):
            return None

        path = request.path

        if self.path is not None:
            path = path[len(self.path) :]

        path = unquote(path.strip('/'))

        if path:
            location = os.path.abspath(os.path.join(self.docroot, path))
        else:
            location = os.path.abspath(os.path.join(self.docroot, '.'))

        if not os.path.exists(location):
            return None

        if not location.startswith(os.path.dirname(self.docroot)):
            return None  # hacking attempt e.g. /foo/../../../../../etc/shadow

        # Is it a file we can serve directly?
        if os.path.isfile(location):
            # Don't set cookies for static content
            response.cookie.clear()
            try:
                return serve_file(request, response, location)
            finally:
                event.stop()

        # Is it a directory?
        elif os.path.isdir(location):
            # Try to serve one of default files first..
            for default in self.defaults:
                location = os.path.abspath(
                    os.path.join(self.docroot, path, default),
                )
                if os.path.exists(location):
                    # Don't set cookies for static content
                    response.cookie.clear()
                    try:
                        return serve_file(request, response, location)
                    finally:
                        event.stop()

            # .. serve a directory listing if allowed to.
            if self.dirlisting:
                directory = os.path.abspath(os.path.join(self.docroot, path))
                cur_dir = os.path.join(self.path, path) if self.path else ''

                if not path:
                    url_up = ''
                else:
                    url_up = os.path.join('/', os.path.split(path)[0]) if self.path is None else os.path.join(cur_dir, '..')
                    url_up = '<li><a href="%s">%s</a></li>' % (escape(url_up, True), '..')

                listing = []
                for item in os.listdir(directory):
                    if not item.startswith('.'):
                        url = os.path.join('/', path, cur_dir, item)
                        location = os.path.abspath(
                            os.path.join(self.docroot, path, item),
                        )
                        if os.path.isdir(location):
                            li = '<li><a href="%s/">%s/</a></li>' % (
                                escape(quote(url), True),
                                escape(item),
                            )
                        else:
                            li = '<li><a href="%s">%s</a></li>' % (
                                escape(quote(url), True),
                                escape(item),
                            )
                        listing.append(li)

                ctx = {}
                ctx['directory'] = cur_dir or os.path.join('/', cur_dir, path)
                ctx['url_up'] = url_up
                ctx['listing'] = '\n'.join(listing)
                try:
                    return _dirlisting_template.safe_substitute(ctx)
                finally:
                    event.stop()
            return None
        return None



#!/usr/bin/env python
import pytest
from hello import App, hello


@pytest.fixture()
def app(request, manager, watcher):
    app = App().register(manager)
    watcher.wait('registered')

    def finalizer():
        app.unregister()

    request.addfinalizer(finalizer)

    return app


def test(app, watcher):
    x = app.fire(hello())
    assert watcher.wait('hello')

    assert x.value == 'Hello World!'



"""stompest StompFrameTransport allowing for ssl.wrap_socket"""

import logging
import socket
import ssl


try:
    from stompest.error import StompConnectionError
    from stompest.sync.transport import StompFrameTransport
except ImportError:
    raise ImportError('No stomp support available. Is stompest installed?')

LOG = logging.getLogger(__name__)


class EnhancedStompFrameTransport(StompFrameTransport):
    """add support for older ssl module and http proxy"""

    proxy_host = None
    proxy_port = None
    proxy_user = None
    proxy_password = None

    def connect(self, timeout=None):
        """Allow older versions of ssl module, allow http proxy connections"""
        LOG.debug('stomp_transport.connect()')
        ssl_params = None
        if isinstance(self.sslContext, dict):
            # This is actually a dictionary of ssl parameters for wrapping the socket
            ssl_params = self.sslContext
            self.sslContext = ssl.create_default_context()
            if ssl_params['ca_certs']:
                self.sslContext.load_verify_locations(cafile=ssl_params['ca_certs'])
            self.sslContext.load_cert_chain(certfile=ssl_params['cert_file'], keyfile=ssl_params['key_file'])
            cert_required = ssl.CERT_REQUIRED if ssl_params['ca_certs'] else ssl.CERT_NONE
            self.sslContext.verify_mode = cert_required
            self.sslContext.minimum_version = ssl_params['ssl_version']

        try:
            if self.proxy_host:
                try:
                    # Don't try to import this unless we need it
                    import socks
                except ImportError:
                    raise ImportError('No http proxy support available. Is pysocks installed?')

                LOG.info('Connecting through proxy %s', self.proxy_host)
                self._socket = socks.socksocket()
                self._socket.set_proxy(
                    socks.HTTP, self.proxy_host, self.proxy_port, True, username=self.proxy_user, password=self.proxy_password
                )
            else:
                self._socket = socket.socket()

            self._socket.settimeout(timeout)
            self._socket.connect((self.host, self.port))

            if self.sslContext:
                self._socket = self.sslContext.wrap_socket(self._socket, server_hostname=self.host)

        except OSError as e:
            raise StompConnectionError('Could not establish connection [%s]' % e)
        self._parser.reset()



#!/usr/bin/env python
import sys

from circuits import Component


class App(Component):
    def started(self, *args):
        try:
            code = int(sys.argv[1])
        except ValueError:
            code = sys.argv[1]

        raise SystemExit(code)


def main():
    App().run()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
from circuits import Component, Event
from circuits.web import Controller

from .helpers import urlopen


class hello(Event):
    """hello Event"""


class App(Component):
    def hello(self):
        return 'Hello World!'


class Root(Controller):
    def index(self):
        return self.fire(hello())


def test(webapp):
    App().register(webapp)

    f = urlopen(webapp.server.http.base)
    s = f.read()
    assert s == b'Hello World!'



#!/usr/bin/env python
from circuits import Component, Event


class test(Event):
    """test Event"""


class hello(Event):
    """hello Event"""


class App(Component):
    def test(self):
        def f():
            while True:
                yield 'Hello'

        return f()

    def hello(self):
        yield 'Hello '
        yield 'World!'


def test_return_generator():
    app = App()
    while len(app):
        app.flush()

    v = app.fire(test())
    app.tick()
    app.tick()

    x = v.value
    assert x == 'Hello'


def test_yield():
    app = App()
    while len(app):
        app.flush()

    v = app.fire(hello())
    app.tick()
    app.tick()
    app.tick()

    x = v.value
    assert x == ['Hello ', 'World!']



import pytest

from .helpers import HTTPError, urlopen


def application(environ, start_response):
    status = '200 OK'
    response_headers = [('Content-type', 'text/plain')]
    start_response(status, response_headers)
    raise Exception('Hello World!')


def test(webapp):
    with pytest.raises(HTTPError) as exc:
        urlopen(webapp.server.http.base)
    assert exc.value.code == 500
    assert exc.value.msg == 'Internal Server Error'
    s = exc.value.read()
    assert b'Exception' in s
    assert b'Hello World!' in s



#!/usr/bin/env python
"""
File Upload

A simple example showing how to access an uploaded file.
"""

from circuits.web import Controller, Server


UPLOAD_FORM = """
<html>
 <head>
  <title>Upload Form</title>
 </head>
 <body>
  <h1>Upload Form</h1>
  <form method="POST" action="/" enctype="multipart/form-data">
   Description: <input type="text" name="desc"><br>
   <input type="file" name="file">
   <input type="submit" value="Submit">
  </form>
 </body>
</html>
"""

UPLOADED_FILE = """
<html>
 <head>
  <title>Uploaded File</title>
 </head>
 <body>
  <h1>Uploaded File</h1>
  <p>
   Filename: %s<br>
   Description: %s
  </p>
  <p><b>File Contents:</b></p>
  <pre>
  %s
  </pre>
 </body>
</html>
"""


class Root(Controller):
    def index(self, file=None, desc=''):
        """
        Request Handler

        If we haven't received an uploaded file yet, repond with
        the UPLOAD_FORM template. Otherwise respond with the
        UPLOADED_FILE template. The file is accessed through
        the ``file`` keyword argument and the description via
        the ``desc`` keyword argument. These also happen to be
        the same fields used on the form.
        """
        if file is None:
            return UPLOAD_FORM
        return UPLOADED_FILE % (file.filename, desc, file.value)


app = Server(('0.0.0.0', 8000))
Root().register(app)
app.run()



#!/usr/bin/env python
"""
An implementation of the Python Concurrency Problem of 99 Bottles of Beer

See: http://wiki.python.org/moin/Concurrency/99Bottles
"""

import sys

from circuits import Component
from circuits.io import File
from circuits.protocols import Line


class Tail(Component):
    """
    A complex component which combines the ``File`` and ``LP``
    (Line Protoco) components together to implement similar functionality to
    the UNIX ``tail`` command.
    """

    def init(self, filename):
        """
        Initialize the Component.

        NB: This is automatically called after ``__new__`` and ``__init__``.
        """
        (File(filename, 'r') + Line()).register(self)


class Grep(Component):
    """
    A simple component that simply listens for ``line`` events from the
    ``Tail`` component and performs a regular expression match against each
    line. If the line matches it is printed to standard output.
    """

    def init(self, pattern):
        self.pattern = pattern

    def line(self, line):
        """Line Event Handler"""
        line = line.decode('utf-8')

        if self.pattern in line:
            print(line)


# Configure and "run" the System.
app = Tail(sys.argv[1])
Grep(sys.argv[2]).register(app)
app.run()



"""Internet Relay Chat Protocol"""

from re import compile as compile_regex

from circuits import Component
from circuits.net.events import write
from circuits.protocols.line import Line

from .commands import PONG
from .events import response
from .utils import parsemsg


NUMERIC = compile_regex('[0-9]+')


class IRC(Component):
    """
    IRC Protocol Component

    Creates a new IRC Component instance that implements the IRC Protocol.
    Incoming messages are handled by the "read" Event Handler, parsed and
    processed with appropriate Events created and exposed to the rest of
    the system to listen to and handle.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.encoding = kwargs.get('encoding', 'utf-8')

        Line(**kwargs).register(self)

    def line(self, *args):
        """
        line Event Handler

        Process a line of text and generate the appropriate
        event. This must not be overridden by subclasses,
        if it is, this must be explicitly called by the
        subclass. Other Components may however listen to
        this event and process custom IRC events.
        """
        if len(args) == 1:
            # Client read
            sock, line = None, args[0]
        else:
            # Server read
            sock, line = args

        prefix, command, args = parsemsg(line, encoding=self.encoding)

        command = command.lower()

        if NUMERIC.match(command):
            args.insert(0, int(command))
            command = 'numeric'

        if sock is not None:
            self.fire(response.create(command, sock, prefix, *args))
        else:
            self.fire(response.create(command, prefix, *args))

    def request(self, event, message):
        """
        request Event Handler (Default)

        This is a default event handler to respond to ``request`` events
        by converting the given message to bytes and firing a ``write``
        event to a hopefully connected client socket.
        Components may override this, but be sure to respond to
        ``request`` events by either explicitly calling this method
        or sending your own ``write`` events as the client socket.
        """
        event.stop()
        message.encoding = self.encoding
        self.fire(write(bytes(message)))

    def ping(self, event, *args):
        """
        ping Event Handler (Default)

        This is a default event to respond to ``ping`` events
        by sending a ``PONG`` in response. Subclasses or
        components may override this, but be sure to respond to
        ``ping`` events by either explicitly calling this method
        or sending your own ``PONG`` response.
        """
        if len(args) == 2:
            # Client read
            self.fire(PONG(args[1]))
            event.stop()



#!/usr/bin/env python
"""
Shadow Auth Demo

An example of a Circuits Component that requires users authenticate
against /etc/passwd or /etc/shadow before letting them into the web site.

curl -i http://test:test@localhost:8000/
curl -i http://root:test@localhost:8000/
"""

from crypt import crypt
from os import path
from re import compile as compile_regex
from socket import gethostname

from circuits import Component, handler
from circuits.web import Controller, Server, _httpauth
from circuits.web.errors import httperror, unauthorized


def check_auth(user, password):
    salt_pattern = compile_regex(r'\$.*\$.*\$')
    passwd = './etc/shadow' if path.exists('./etc/shadow') else '/etc/passwd'

    with open(passwd) as f:
        rows = (line.strip().split(':') for line in f)
        records = [row for row in rows if row[0] == user]

    if not records:
        return False
    hash = records[0][1]
    salt = salt_pattern.match(hash).group()

    return crypt(password, salt) == hash


class PasswdAuth(Component):
    channel = 'web'

    def init(self, realm=None):
        self.realm = realm or gethostname()

    @handler('request', priority=1.0)
    def _on_request(self, event, request, response):
        if 'authorization' in request.headers:
            ah = _httpauth.parseAuthorization(request.headers['authorization'])
            if ah is None:
                event.stop()
                return httperror(request, response, 400)

            username, password = ah['username'], ah['password']

            if check_auth(username, password):
                request.login = username
                return None

        response.headers['WWW-Authenticate'] = _httpauth.basicAuth(self.realm)

        event.stop()
        return unauthorized(request, response)


class Root(Controller):
    def index(self):
        return f'Hello, {self.request.login:s}'


app = Server(('0.0.0.0', 8000))
PasswdAuth().register(app)
Root().register(app)
app.run()



#!/usr/bin/env python
import sys
from types import ModuleType

from circuits import Component
from circuits.core.utils import findchannel, findroot, findtype


FOO = """\
def foo():
    return "Hello World!"
"""

FOOBAR = """\
def foo();
    return "Hello World!'
"""


class Base(Component):
    """Base"""


class App(Base):
    def hello(self):
        return 'Hello World!'


class A(Component):
    channel = 'a'


class B(Component):
    channel = 'b'


def test_safeimport(tmpdir):
    from circuits.core.utils import safeimport

    sys.path.insert(0, str(tmpdir))

    foo_path = tmpdir.ensure('foo.py')
    foo_path.write(FOO)

    foo = safeimport('foo')
    assert foo is not None
    assert type(foo) is ModuleType

    s = foo.foo()
    assert s == 'Hello World!'
    pyc = foo_path.new(ext='pyc')
    if pyc.check(file=1):
        pyc.remove(ignore_errors=True)
    pyd = foo_path.dirpath('__pycache__')
    if pyd.check(dir=1):
        pyd.remove(ignore_errors=True)
    foo_path.write(FOOBAR)

    foo = safeimport('foo')
    assert foo is None
    assert foo not in sys.modules


def test_findroot():
    app = App()
    a = A()
    b = B()

    b.register(a)
    a.register(app)

    while len(app):
        app.flush()

    root = findroot(b)

    assert root == app


def test_findchannel():
    app = App()
    (A() + B()).register(app)

    while len(app):
        app.flush()

    a = findchannel(app, 'a')

    assert a.channel == 'a'


def test_findtype():
    app = App()
    (A() + B()).register(app)

    while len(app):
        app.flush()

    a = findtype(app, A)

    assert isinstance(a, A)



#!/usr/bin/python -i
from os import getpid

import pytest

from circuits import Component, Event, ipc


pytestmark = pytest.mark.skipif(pytest.PLATFORM == 'win32', reason='Unsupported Platform')

pytest.importorskip('multiprocessing')


class hello(Event):
    """hello Event"""


class App(Component):
    def hello(self):
        return f'Hello from {getpid():d}'


def test(manager, watcher):
    app = App()
    _process, bridge = app.start(process=True, link=manager)
    assert watcher.wait('ready')

    x = manager.fire(ipc(hello()))

    assert pytest.wait_for(x, 'result')

    assert x.value == f'Hello from {app.pid:d}'

    app.stop()
    app.join()

    bridge.unregister()
    watcher.wait('unregistered')



#!/usr/bin/env python
import os

import mako
from mako.lookup import TemplateLookup

from circuits.web import Controller, Server, Static


DEFAULTS = {}

templates = TemplateLookup(
    directories=[os.path.join(os.path.dirname(__file__), 'tpl')],
    module_directory='/tmp',
    output_encoding='utf-8',
)


def render(name, **d):
    try:
        d.update(DEFAULTS)
        tpl = templates.get_template(name)
        return tpl.render(**d)
    except Exception:
        return mako.exceptions.html_error_template().render()


class Root(Controller):
    tpl = 'index.html'

    def index(self):
        return render(self.tpl)

    def submit(self, firstName, lastName):
        msg = f'Thank you {firstName} {lastName}'
        return render(self.tpl, message=msg)


app = Server(('0.0.0.0', 8000))
Static().register(app)
Root().register(app)
app.run()



#!/usr/bin/env python
import pytest

from circuits import Component, Event


class test(Event):
    """test Event"""


class App(Component):
    def __init__(self):
        super().__init__()

        self.etype = None
        self.evalue = None
        self.etraceback = None
        self.handler = None
        self.fevent = None

    def test(self):
        return x  # noqa: F821

    def exception(self, etype, evalue, etraceback, handler=None, fevent=None):
        self.etype = etype
        self.evalue = evalue
        self.etraceback = etraceback
        self.handler = handler
        self.fevent = fevent


def reraise(e):
    raise e


@pytest.fixture()
def app(request, manager, watcher):
    app = App().register(manager)
    watcher.wait('registered')

    def finalizer():
        app.unregister()

    request.addfinalizer(finalizer)

    return app


def test_main(app, watcher):
    e = test()
    app.fire(e)
    watcher.wait('exception')

    assert app.etype is NameError
    pytest.raises(NameError, lambda e: reraise(e), app.evalue)
    assert isinstance(app.etraceback, list)
    assert app.handler == app.test
    assert app.fevent == e



#!/usr/bin/env python
from circuits.web import Controller

from .helpers import CookieJar, HTTPCookieProcessor, build_opener


class Root(Controller):
    def index(self):
        visited = self.cookie.get('visited')
        if visited and visited.value:
            return 'Hello again!'
        self.cookie['visited'] = True
        return 'Hello World!'


def test(webapp):
    cj = CookieJar()
    opener = build_opener(HTTPCookieProcessor(cj))

    f = opener.open(webapp.server.http.base)
    s = f.read()
    assert s == b'Hello World!'

    f = opener.open(webapp.server.http.base)
    s = f.read()
    assert s == b'Hello again!'



#!/usr/bin/python -i
from types import TracebackType

import pytest

from circuits import Component, Event, handler


class hello(Event):
    """Hello Event"""


class test(Event):
    """test Event"""


class foo(Event):
    """foo Event"""


class values(Event):
    """values Event"""

    complete = True


class App(Component):
    def hello(self):
        return 'Hello World!'

    def test(self):
        return self.fire(hello())

    def foo(self):
        raise Exception('ERROR')

    @handler('hello_value_changed')
    def _on_hello_value_changed(self, value):
        self.value = value

    @handler('test_value_changed')
    def _on_test_value_changed(self, value):
        self.value = value

    @handler('values', priority=2.0)
    def _value1(self):
        return 'foo'

    @handler('values', priority=1.0)
    def _value2(self):
        return 'bar'

    @handler('values', priority=0.0)
    def _value3(self):
        return self.fire(hello())


@pytest.fixture()
def app(request, simple_manager):
    return App().register(simple_manager)


def test_value(app, simple_manager):
    x = app.fire(hello())
    simple_manager.run()

    assert 'Hello World!' in x
    assert x.value == 'Hello World!'


def test_nested_value(app, simple_manager):
    x = app.fire(test())
    simple_manager.run()

    assert x.value == 'Hello World!'
    assert str(x) == 'Hello World!'


def test_value_notify(app, simple_manager):
    ev = hello()
    ev.notify = True
    x = app.fire(ev)

    simple_manager.run()

    assert 'Hello World!' in x
    assert x.value == 'Hello World!'
    assert app.value is x


def test_nested_value_notify(app, simple_manager):
    ev = test()
    ev.notify = True
    x = app.fire(ev)

    simple_manager.run()

    assert x.value == 'Hello World!'
    assert str(x) == 'Hello World!'
    assert app.value is x


def test_error_value(app, simple_manager):
    x = app.fire(foo())
    simple_manager.run()

    etype, evalue, etraceback = x
    assert etype is Exception
    assert str(evalue) == 'ERROR'
    assert isinstance(etraceback, TracebackType)


def test_multiple_values(app, simple_manager):
    v = app.fire(values())
    simple_manager.run()

    assert isinstance(v.value, list)

    x = list(v)

    assert 'foo' in v
    assert x == ['foo', 'bar', 'Hello World!']
    assert x[0] == 'foo'
    assert x[1] == 'bar'
    assert x[2] == 'Hello World!'



#!/usr/bin/env python

from time import sleep

from circuits import Component, Debugger, Event, Timer, Worker, task


def factorial(n):
    x = 1
    for i in range(1, (n + 1)):
        x = x * (i + 1)
        sleep(1)  # deliberate!
    return x


class App(Component):
    def init(self, *args, **kwargs):
        Worker(process=True).register(self)

    def foo(self):
        print('Foo!')

    def started(self, component):
        self.fire(task(factorial, 3))  # async
        self.fire(task(factorial, 5))  # async
        self.fire(task(factorial, 7))  # async
        self.fire(task(factorial, 10))  # async
        self.fire(task(factorial, 11))  # async
        self.fire(task(factorial, 11))  # async
        self.fire(task(factorial, 12))  # async
        self.fire(task(factorial, 14))  # async
        Timer(1, Event.create('foo'), persist=True).register(self)

    def task_success(self, function_called, factorial_result):
        _func, argument = function_called
        print(f'factorial({argument!s}) = {factorial_result:d}')
        # Stop after the last and longest running task
        if argument == 14:
            self.stop()


if __name__ == '__main__':
    app = App()
    Debugger().register(app)
    app.run()



#!/usr/bin/env python

from circuits import Component
from circuits.tools import graph


class Pound(Component):
    def __init__(self):
        super().__init__()

        self.bob = Bob().register(self)
        self.fred = Fred().register(self)

    def started(self, *args):
        print(graph(self.root))


class Bob(Component):
    def started(self, *args):
        print("Hello I'm Bob!")


class Fred(Component):
    def started(self, *args):
        print("Hello I'm Fred!")


Pound().run()



#!/usr/bin/env python
from circuits.web import Controller, Server
from circuits.web.tools import basic_auth, check_auth


class Root(Controller):
    def index(self):
        realm = 'Test'
        users = {'admin': 'admin'}
        encrypt = str

        if check_auth(self.request, self.response, realm, users, encrypt):
            return 'Hello %s' % self.request.login

        return basic_auth(self.request, self.response, realm, users, encrypt)


app = Server(('0.0.0.0', 8000))
Root().register(app)
app.run()



#!/usr/bin/env python
from circuits import Debugger
from circuits.web import Controller, Server


class Root(Controller):
    def index(self):
        return 'Hello World!'


app = Server(('0.0.0.0', 8443), secure=True, certfile='cert.pem')
Debugger().register(app)
Root().register(app)
app.run()



#!/usr/bin/env python
"""
Node Server Example

This example demonstrates how to create a very simple node server
that supports bi-diractional messaging between server and connected
clients forming a cluster of nodes.
"""

from optparse import OptionParser
from os import getpid

from circuits import Component, Debugger
from circuits.node import Node


__version__ = '0.0.1'

USAGE = '%prog [options]'
VERSION = '%prog v' + __version__


def parse_options():
    parser = OptionParser(usage=USAGE, version=VERSION)

    parser.add_option(
        '-b',
        '--bind',
        action='store',
        type='string',
        default='0.0.0.0:8000',
        dest='bind',
        help='Bind to address:[port]',
    )

    parser.add_option(
        '-d',
        '--debug',
        action='store_true',
        default=False,
        dest='debug',
        help='Enable debug mode',
    )

    opts, args = parser.parse_args()

    return opts, args


class NodeServer(Component):
    def init(self, args, opts):
        """
        Initialize our ``ChatServer`` Component.

        This uses the convenience ``init`` method which is called after the
        component is proeprly constructed and initialized and passed the
        same args and kwargs that were passed during construction.
        """
        self.args = args
        self.opts = opts

        self.clients = {}

        if opts.debug:
            Debugger().register(self)

        if ':' in opts.bind:
            address, port = opts.bind.split(':')
            port = int(port)
        else:
            address, port = opts.bind, 8000

        Node(port=port, server_ip=address).register(self)

    def connect(self, sock, host, port):
        """Connect Event -- Triggered for new connecting clients"""
        self.clients[sock] = {
            'host': sock,
            'port': port,
        }

    def disconnect(self, sock):
        """Disconnect Event -- Triggered for disconnecting clients"""
        if sock not in self.clients:
            return

        del self.clients[sock]

    def ready(self, server, bind):
        print('Ready! Listening on {}:{}'.format(*bind))
        print('Waiting for remote events...')

    def hello(self):
        return f'Hello World! ({getpid():d})'


def main():
    opts, args = parse_options()

    # Configure and "run" the System.
    NodeServer(args, opts).run()


if __name__ == '__main__':
    main()



"""This module defines the Manager class."""

import _thread
import atexit
import contextlib
import types
from collections import deque
from heapq import heappop, heappush
from inspect import isfunction
from itertools import chain, count
from multiprocessing import Process, current_process
from operator import attrgetter
from os import getpid, kill
from signal import SIGINT, SIGTERM, signal as set_signal_handler
from sys import exc_info as _exc_info, stderr
from threading import RLock, Thread, current_thread
from time import time
from traceback import format_exc
from types import GeneratorType
from uuid import uuid4 as uuid

from .events import Event, exception, generate_events, signal, started, stopped
from .handlers import handler
from .values import Value


try:
    from signal import SIGKILL
except ImportError:
    SIGKILL = SIGTERM


TIMEOUT = 0.1  # 100ms timeout when idle


class UnregistrableError(Exception):
    """Raised if a component cannot be registered as child."""


class TimeoutError(Exception):
    """Raised if wait event timeout occurred"""


class CallValue:
    def __init__(self, value):
        self.value = value


class ExceptionWrapper:
    def __init__(self, exception):
        self.exception = exception

    def extract(self):
        return self.exception


class Sleep:
    def __init__(self, seconds):
        self._task = None

        try:
            self.expiry = time() + float(seconds)
        except ValueError:
            raise TypeError('a float is required')

    def __iter__(self):
        return self

    def __repr__(self):
        return f'sleep({self.expiry - time()!r})'

    def __next__(self):
        if time() >= self.expiry:
            raise StopIteration()
        return self

    @property
    def expired(self):
        return time() >= self.expiry

    @property
    def task(self):
        return self._task

    @task.setter
    def task(self, task):
        self._task = task


def sleep(seconds):
    """
    Delay execution of a coroutine for a given number of seconds.
    The argument may be a floating point number for subsecond precision.
    """
    return Sleep(seconds)


class Dummy:
    channel = None


_dummy = Dummy()
del Dummy


class _State:
    __slots__ = ('event', 'flag', 'parent', 'run', 'task', 'task_event', 'tick_handler', 'timeout')

    def __init__(self, timeout):
        self.task = None
        self.run = False
        self.flag = False
        self.event = None
        self.timeout = timeout
        self.parent = None
        self.task_event = None
        self.tick_handler = None


class _EventQueue:
    __slots__ = ('_counter', '_flush_batch', '_priority_queue', '_queue')

    def __init__(self):
        self._queue = deque()
        self._priority_queue = []
        self._counter = count()
        self._flush_batch = 0

    def __len__(self):
        return len(self._queue) + len(self._priority_queue)

    def drainFrom(self, other_queue):
        self._queue.extend(other_queue._queue)
        other_queue._queue.clear()
        # Queue is currently flushing events /o\
        assert not len(other_queue._priority_queue)

    def append(self, event, channel, priority):
        self._queue.append((priority, next(self._counter), (event, channel)))

    def dispatchEvents(self, dispatcher):
        if self._flush_batch == 0:
            # FIXME: Might be faster to use heapify instead of pop +
            # heappush. Though, with regards to thread safety this
            # appears to be the better approach.
            self._flush_batch = count = len(self._queue)
            while count:
                count -= 1
                heappush(self._priority_queue, self._queue.popleft())

        while self._flush_batch > 0:
            self._flush_batch -= 1  # Decrement first!
            (event, channels) = heappop(self._priority_queue)[2]
            dispatcher(event, channels, self._flush_batch)


class Manager:
    """
    The manager class has two roles. As a base class for component
    implementation, it provides methods for event and handler management.
    The method :meth:`.fireEvent` appends a new event at the end of the event
    queue for later execution. :meth:`.waitEvent` suspends the execution
    of a handler until all handlers for a given event have been invoked.
    :meth:`.callEvent` combines the last two methods in a single method.

    The methods :meth:`.addHandler` and :meth:`.removeHandler` allow handlers
    for events to be added and removed dynamically. (The more common way to
    register a handler is to use the :func:`~.handlers.handler` decorator
    or derive the class from :class:`~.components.Component`.)

    In its second role, the :class:`.Manager` takes the role of the
    event executor. Every component hierarchy has a root component that
    maintains a queue of events. Firing an event effectively means
    appending it to the event queue maintained by the root manager.
    The :meth:`.flush` method removes all pending events from the
    queue and, for each event, invokes all the handlers. Usually,
    :meth:`.flush` is indirectly invoked by :meth:`run`.

    The manager optionally provides information about the execution of
    events as automatically generated events. If an :class:`~.events.Event`
    has its :attr:`success` attribute set to True, the manager fires
    a :class:`~.events.Success` event if all handlers have been
    executed without error. Note that this event will be
    enqueued (and dispatched) immediately after the events that have been
    fired by the event's handlers. So the success event indicates both
    the successful invocation of all handlers for the event and the
    processing of the immediate follow-up events fired by those handlers.

    Sometimes it is not sufficient to know that an event and its
    immediate follow-up events have been processed. Rather, it is
    important to know when all state changes triggered by an event,
    directly or indirectly, have been performed. This also includes
    the processing of events that have been fired when invoking
    the handlers for the follow-up events and the processing of events
    that have again been fired by those handlers and so on. The completion
    of the processing of an event and all its direct or indirect
    follow-up events may be indicated by a :class:`~.events.Complete`
    event. This event is generated by the manager if :class:`~.events.Event`
    has its :attr:`complete` attribute set to True.

    Apart from the event queue, the root manager also maintains a list of
    tasks, actually Python generators, that are updated when the event queue
    has been flushed.
    """

    _currently_handling = None
    """
    The event currently being handled.
    """

    def __init__(self, *args, **kwargs):
        """Initializes x; see x.__class__.__doc__ for signature"""
        self._queue = _EventQueue()

        self._tasks = set()
        self._cache = {}
        self._globals = set()
        self._handlers = {}

        self._flush_batch = 0
        self._cache_needs_refresh = False

        self._executing_thread = None
        self._flushing_thread = None
        self._running = False
        self.__thread = None
        self.__process = None
        self._lock = RLock()

        self.root = self.parent = self
        self.components = set()

    def __nonzero__(self):
        """x.__nonzero__() <==> bool(x)"""
        return True

    __bool__ = __nonzero__

    def __repr__(self):
        """x.__repr__() <==> repr(x)"""
        name = self.__class__.__name__

        channel = '/{}'.format(getattr(self, 'channel', ''))

        q = len(self._queue)
        state = 'R' if self.running else 'S'

        pid = current_process().pid

        id = f'{pid}:{current_thread().name}' if pid else current_thread().name

        format = '<%s%s %s (queued=%d) [%s]>'
        return format % (name, channel, id, q, state)

    def __contains__(self, y):
        """
        x.__contains__(y) <==> y in x

        Return True if the Component y is registered.
        """
        components = self.components.copy()
        return y in components or y in [c.__class__ for c in components]

    def __len__(self):
        """
        x.__len__() <==> len(x)

        Returns the number of events in the Event Queue.
        """
        return len(self._queue)

    def __add__(self, y):
        """
        x.__add__(y) <==> x+y

        (Optional) Convenience operator to register y with x
        Equivalent to: y.register(x)

        @return: x
        @rtype Component or Manager
        """
        y.register(self)
        return self

    def __iadd__(self, y):
        """
        x.__iadd__(y) <==> x += y

        (Optional) Convenience operator to register y with x
        Equivalent to: y.register(x)

        @return: x
        @rtype Component or Manager
        """
        y.register(self)
        return self

    def __sub__(self, y):
        """
        x.__sub__(y) <==> x-y

        (Optional) Convenience operator to unregister y from x.parent
        Equivalent to: y.unregister()

        @return: x
        @rtype Component or Manager
        """
        if y.parent is not y:
            y.unregister()
        return self

    def __isub__(self, y):
        """
        x.__sub__(y) <==> x -= y

        (Optional) Convenience operator to unregister y from x
        Equivalent to: y.unregister()

        @return: x
        @rtype Component or Manager
        """
        if y.parent is not y:
            y.unregister()
        return self

    @property
    def name(self):
        """Return the name of this Component/Manager"""
        return self.__class__.__name__

    @property
    def running(self):
        """Return the running state of this Component/Manager"""
        return self._running

    @property
    def pid(self):
        """Return the process id of this Component/Manager"""
        return getpid() if self.__process is None else self.__process.pid

    def getHandlers(self, event, channel, **kwargs):
        name = event.name
        handlers = set()

        _handlers = set()
        _handlers.update(self._handlers.get('*', []))
        _handlers.update(self._handlers.get(name, []))

        for _handler in _handlers:
            handler_channel = _handler.channel
            if handler_channel is None:
                # TODO: Why do we care about the event handler's channel?
                #      This probably costs us performance for what?
                #      I've not ever had to rely on this in practice...
                handler_channel = getattr(
                    getattr(_handler, 'im_self', getattr(_handler, '__self__', _dummy)),
                    'channel',
                    None,
                )

            if channel == '*' or handler_channel in ('*', channel) or channel is self:
                handlers.add(_handler)

        if not kwargs.get('exclude_globals', False):
            handlers.update(self._globals)

        for c in self.components.copy():
            handlers.update(c.getHandlers(event, channel, **kwargs))

        return handlers

    def addHandler(self, f):
        method = types.MethodType(f, self) if isfunction(f) else f

        setattr(self, method.__name__, method)

        if not method.names and method.channel == '*':
            self._globals.add(method)
        elif not method.names:
            self._handlers.setdefault('*', set()).add(method)
        else:
            for name in method.names:
                self._handlers.setdefault(name, set()).add(method)

        self.root._cache_needs_refresh = True

        return method

    def removeHandler(self, method, event=None):
        names = method.names if event is None else [event]

        for name in names:
            self._handlers[name].remove(method)
            if not self._handlers[name]:
                del self._handlers[name]
                try:
                    delattr(self, method.__name__)
                except AttributeError:
                    # Handler was never part of self
                    pass

        self.root._cache_needs_refresh = True

    def registerChild(self, component):
        if component._executing_thread is not None:
            if self.root._executing_thread is not None:
                raise UnregistrableError()
            self.root._executing_thread = component._executing_thread
            component._executing_thread = None
        self.components.add(component)
        self.root._queue.drainFrom(component._queue)
        self.root._cache_needs_refresh = True

    def unregisterChild(self, component):
        self.components.remove(component)
        self.root._cache_needs_refresh = True

    def _fire(self, event, channel, priority=0):
        # check if event is fired while handling an event
        th = self._executing_thread or self._flushing_thread
        if _thread.get_ident() == (th.ident if th else None) and not isinstance(event, signal):
            if self._currently_handling is not None and getattr(self._currently_handling, 'cause', None):
                # if the currently handled event wants to track the
                # events generated by it, do the tracking now
                event.cause = self._currently_handling
                event.effects = 1
                self._currently_handling.effects += 1

            self._queue.append(event, channel, priority)

        # the event comes from another thread
        else:
            # Another thread has provided us with something to do.
            # If the component is running, we must make sure that
            # any pending generate event waits no longer, as there
            # is something to do now.
            with self._lock:
                # Modifications of attribute self._currently_handling
                # (in _dispatch()), calling reduce_time_left(0). and adding an
                # event to the (empty) event queue must be atomic, so we have
                # to lock. We can save the locking around
                # self._currently_handling = None though, but then need to copy
                # it to a local variable here before performing a sequence of
                # operations that assume its value to remain unchanged.
                handling = self._currently_handling

                self._queue.append(event, channel, priority)
                if isinstance(handling, generate_events):
                    handling.reduce_time_left(0)

    def fireEvent(self, event, *channels, **kwargs):
        """
        Fire an event into the system.

        :param event: The event that is to be fired.
        :param channels: The channels that this event is delivered on.
           If no channels are specified, the event is delivered to the
           channels found in the event's :attr:`channel` attribute.
           If this attribute is not set, the event is delivered to
           the firing component's channel. And eventually,
           when set neither, the event is delivered on all
           channels ("*").
        """
        if not channels:
            channels = event.channels or (getattr(self, 'channel', '*'),)

        event.channels = channels

        event.value = Value(event, self)
        self.root._fire(event, channels, **kwargs)

        return event.value

    fire = fireEvent

    def registerTask(self, g):
        self.root._tasks.add(g)

    def unregisterTask(self, g):
        if g in self.root._tasks:
            self.root._tasks.remove(g)

    def waitEvent(self, event, *channels, **kwargs):  # noqa
        # TODO: C901: This has a high McCabe complexity score of 16.
        # TODO: Refactor this method.

        if isinstance(event, Event):
            event_object = event
            event_name = event.name
            channels = event.channels or channels
        else:
            event_object = None
            event_name = event

        state = _State(timeout=kwargs.get('timeout', -1))

        def _on_event(self, event, *args, **kwargs):
            if not state.run and (event_object is None or event is event_object):
                self.removeHandler(_on_event_handler, event_name)
                event.alert_done = True
                state.run = True
                state.event = event

        def _on_done(self, event, *args, **kwargs):
            if state.event == event.parent:
                state.flag = True
                self.registerTask((state.task_event, state.task, state.parent))
                if state.timeout > 0:
                    self.removeHandler(state.tick_handler, 'generate_events')

        def _on_tick(self):
            if state.timeout == 0:
                self.registerTask(
                    (
                        state.task_event,
                        (e for e in (ExceptionWrapper(TimeoutError()),)),
                        state.parent,
                    ),
                )
                self.removeHandler(_on_done_handler, '%s_done' % event_name)
                self.removeHandler(_on_tick_handler, 'generate_events')
            elif state.timeout > 0:
                state.timeout -= 1

        if not channels:
            channels = (None,)

        for channel in channels:
            _on_event_handler = self.addHandler(handler(event_name, channel=channel)(_on_event))
            _on_done_handler = self.addHandler(handler('%s_done' % event_name, channel=channel)(_on_done))
            if state.timeout >= 0:
                _on_tick_handler = state.tick_handler = self.addHandler(handler('generate_events', channel=channel)(_on_tick))

        yield state

        self.removeHandler(_on_done_handler, '%s_done' % event_name)

        if state.event is not None:
            yield CallValue(state.event.value)

    wait = waitEvent

    def callEvent(self, event, *channels, **kwargs):
        """
        Fire the given event to the specified channels and suspend
        execution until it has been dispatched. This method may only
        be invoked as argument to a ``yield`` on the top execution level
        of a handler (e.g. "``yield self.callEvent(event)``").
        It effectively creates and returns a generator
        that will be invoked by the main loop until the event has
        been dispatched (see :func:`circuits.core.handlers.handler`).
        """
        value = self.fire(event, *channels)
        yield from self.waitEvent(event, *event.channels, **kwargs)
        yield CallValue(value)

    call = callEvent

    def _flush(self):
        # Handle events currently on queue, but none of the newly generated
        # events. Note that _flush can be called recursively.
        old_flushing = self._flushing_thread
        try:
            self._flushing_thread = current_thread()
            self._queue.dispatchEvents(self._dispatcher)
        finally:
            self._flushing_thread = old_flushing

    def flushEvents(self):
        """
        Flush all Events in the Event Queue. If called on a manager
        that is not the root of an object hierarchy, the invocation
        is delegated to the root manager.
        """
        self.root._flush()

    flush = flushEvents

    def _dispatcher(self, event, channels, remaining):  # noqa
        # TODO: C901: This has a high McCabe complexity score of 22.
        # TODO: Refactor this method.

        if event.cancelled:
            return

        if event.complete:
            if not getattr(event, 'cause', None):
                event.cause = event
            event.effects = 1  # event itself counts (must be done)
        eargs = event.args
        ekwargs = event.kwargs

        if self._cache_needs_refresh:
            # Don't call self._cache.clear() from other threads,
            # this may interfere with cache rebuild.
            self._cache.clear()
            self._cache_needs_refresh = False
        try:  # try/except is fastest if successful in most cases
            event_handlers = self._cache[(event.name, channels)]
        except KeyError:
            h = (self.getHandlers(event, channel) for channel in channels)

            event_handlers = sorted(
                chain(*h),
                key=attrgetter('priority'),
                reverse=True,
            )

            if isinstance(event, generate_events):
                from .helpers import FallBackGenerator

                event_handlers.append(FallBackGenerator()._on_generate_events)
            elif isinstance(event, exception) and len(event_handlers) == 0:
                from .helpers import FallBackExceptionHandler

                event_handlers.append(FallBackExceptionHandler()._on_exception)
            elif isinstance(event, signal) and len(event_handlers) == 0:
                from .helpers import FallBackSignalHandler

                event_handlers.append(FallBackSignalHandler()._on_signal)

            self._cache[(event.name, channels)] = event_handlers

        if isinstance(event, generate_events):
            with self._lock:
                self._currently_handling = event
                if remaining > 0 or len(self._queue) or not self._running:
                    event.reduce_time_left(0)
                elif self._tasks:
                    event.reduce_time_left(TIMEOUT)
                # From now on, firing an event will reduce time left
                # to 0, which prevents event handlers from waiting (or wakes
                # them up with resume if they should be waiting already)
        else:
            self._currently_handling = event

        value = None
        err = None

        for event_handler in event_handlers:
            event.handler = event_handler
            try:
                value = event_handler(event, *eargs, **ekwargs) if event_handler.event else event_handler(*eargs, **ekwargs)
            except KeyboardInterrupt:
                self.stop()
            except SystemExit as e:
                self.stop(e.code)
            except BaseException:
                value = err = _exc_info()
                event.value.errors = True

                if event.failure:
                    self.fire(event.child('failure', event, err), *event.channels)

                self.fire(exception(*err, handler=event_handler, fevent=event))

            if value is not None:
                if isinstance(value, GeneratorType):
                    event.waitingHandlers += 1
                    event.value.promise = True
                    self.registerTask((event, value, None))
                else:
                    event.value.value = value

            # it is kind of a temporal hack to allow processing
            # of tasks, added in one of event handlers here
            if isinstance(event, generate_events) and self._tasks:
                event.reduce_time_left(TIMEOUT)

            if event.stopped:
                break  # Stop further event processing

        self._currently_handling = None
        self._eventDone(event, err)

    def _eventDone(self, event, err=None):
        if event.waitingHandlers:
            return

        # The "%s_done" event is for internal use by waitEvent only.
        # Use the "%s_success" event in your application if you are
        # interested in being notified about the last handler for
        # an event having been invoked.
        if event.alert_done:
            self.fire(event.child('done', event.value.value), *event.channels)

        if err is None and event.success:
            channels = getattr(event, 'success_channels', event.channels)
            self.fire(event.child('success', event, event.value.value), *channels)

        while True:
            # cause attributes indicates interest in completion event
            cause = getattr(event, 'cause', None)
            if not cause:
                break
            # event takes part in complete detection (as nested or root event)
            event.effects -= 1
            if event.effects > 0:
                break  # some nested events remain to be completed
            if event.complete:  # does this event want signaling?
                self.fire(
                    event.child('complete', event, event.value.value),
                    *getattr(event, 'complete_channels', event.channels),
                )

            # this event and nested events are done now
            delattr(event, 'cause')
            delattr(event, 'effects')
            # cause has one of its nested events done, decrement and check
            event = cause

    def _signal_handler(self, signo, stack):
        self.fire(signal(signo, stack))

    def start(self, process=False, link=None):
        """
        Start a new thread or process that invokes this manager's
        ``run()`` method. The invocation of this method returns
        immediately after the task or process has been started.
        """
        if process:
            # Parent<->Child Bridge
            if link is not None:
                from circuits.core.bridge import Bridge
                from circuits.net.sockets import Pipe

                channels = (uuid(),) * 2
                parent, child = Pipe(*channels)
                bridge = Bridge(parent, channel=channels[0]).register(link)

                args = (child,)
            else:
                args = ()
                bridge = None

            self.__process = Process(target=self.run, args=args, name=self.name)
            self.__process.daemon = True
            self.__process.start()

            return self.__process, bridge
        self.__thread = Thread(target=self.run, name=self.name)
        self.__thread.daemon = True
        self.__thread.start()

        return self.__thread, None

    def join(self):
        if self.__thread is not None:
            return self.__thread.join()

        if self.__process is not None:
            return self.__process.join()
        return None

    def stop(self, code=None):
        """
        Stop this manager. Invoking this method causes
        an invocation of ``run()`` to return.
        """
        if self.__process not in (None, current_process()) and self.__process.is_alive():
            self.__process.terminate()
            self.__process.join(TIMEOUT)

            if self.__process.is_alive():
                kill(self.__process.pid, SIGKILL)

        if not self.running:
            return

        self._running = False

        self.fire(stopped(self))

        if self.root._executing_thread is None:
            for _ in range(3):
                self.tick()

        if code is not None:
            raise SystemExit(code)

    def processTask(self, event, task, parent=None):  # noqa
        # TODO: C901: This has a high McCabe complexity score of 16.
        # TODO: Refactor this method.

        value = None
        try:
            value = next(task)
            if isinstance(value, CallValue):
                # Done here, next() will StopIteration anyway
                self.unregisterTask((event, task, parent))
                # We are in a callEvent
                value = parent.send(value.value)
                if isinstance(value, GeneratorType):
                    # We loose a yield but we gain one,
                    # we don't need to change
                    # event.waitingHandlers
                    # The below code is delegated to handlers
                    # in the waitEvent generator
                    # self.registerTask((event, value, parent))
                    task_state = next(value)
                    task_state.task_event = event
                    task_state.task = value
                    task_state.parent = parent
                else:
                    event.waitingHandlers -= 1
                    if value is not None:
                        event.value.value = value
                    self.registerTask((event, parent, None))
            elif isinstance(value, GeneratorType):
                event.waitingHandlers += 1
                self.unregisterTask((event, task, None))
                # First yielded value is always the task state
                task_state = next(value)
                task_state.task_event = event
                task_state.task = value
                task_state.parent = task
                # The below code is delegated to handlers
                # in the waitEvent generator
                # self.registerTask((event, value, task))
                # TODO: ^^^ Why is this commented out anyway?
            elif isinstance(value, ExceptionWrapper):
                self.unregisterTask((event, task, parent))
                if parent:
                    value = parent.throw(value.extract())
                    if value is not None:
                        value_generator = (val for val in (value,))
                        self.registerTask((event, value_generator, parent))
                else:
                    raise value.extract()
            elif isinstance(value, Sleep):
                if value is not task:
                    value.task = (event, task, parent)
                    self.registerTask((event, value, parent))
                    self.unregisterTask((event, task, parent))
            elif value is not None:
                event.value.value = value
        except StopIteration:
            event.waitingHandlers -= 1
            self.unregisterTask((event, task, parent))

            if parent:
                self.registerTask((event, parent, None))
            elif hasattr(task, 'task'):
                # TODO: The subtask is considered a "waiting handler"
                event.waitingHandlers += 1
                self.registerTask(task.task)
            elif event.waitingHandlers == 0:
                event.value.inform(True)
                self._eventDone(event)
        except KeyboardInterrupt:
            self.stop()
        except SystemExit as e:
            self.stop(e.code)
        except BaseException:
            self.unregisterTask((event, task, parent))

            err = _exc_info()

            event.value.value = err
            event.value.errors = True
            event.value.inform(True)

            if event.failure:
                self.fire(event.child('failure', event, err), *event.channels)

            self.fire(exception(*err, handler=None, fevent=event))

    def tick(self, timeout=-1):
        """
        Execute all possible actions once. Process all registered tasks
        and flush the event queue. If the application is running fire a
        GenerateEvents to get new events from sources.

        This method is usually invoked from :meth:`~.run`. It may also be
        used to build an application specific main loop.

        :param timeout: the maximum waiting time spent in this method. If
            negative, the method may block until at least one action
            has been taken.
        :type timeout: float, measuring seconds
        """
        # process tasks
        if self._tasks:
            for task in self._tasks.copy():
                self.processTask(*task)

        if self._running:
            self.fire(generate_events(self._lock, timeout), '*')

        if len(self._queue):
            self.flush()

    def run(self, socket=None):
        """
        Run this manager. The method fires the
        :class:`~.events.Started` event and then continuously
        calls :meth:`~.tick`.

        The method returns when the manager's
        :meth:`~.stop` method is invoked.

        If invoked by a programs main thread, a signal handler for
        the ``INT`` and ``TERM`` signals is installed. This handler
        fires the corresponding :class:`~.events.Signal`
        events and then calls :meth:`~.stop` for the manager.
        """
        atexit.register(self.stop)

        if current_thread().name == 'MainThread':
            try:
                set_signal_handler(SIGINT, self._signal_handler)
                set_signal_handler(SIGTERM, self._signal_handler)
            except ValueError:
                # Ignore if we can't install signal handlers
                pass

        self._running = True
        self.root._executing_thread = current_thread()

        # Setup Communications Bridge

        if socket is not None:
            from circuits.core.bridge import Bridge

            Bridge(socket, channel=socket.channel).register(self)

        self.fire(started(self))

        try:
            while self.running or len(self._queue):
                self.tick()
            # Fading out, handle remaining work from stop event
            for _ in range(3):
                self.tick()
        except Exception as exc:
            stderr.write(f'Unhandled ERROR: {exc}\n')
            stderr.write(format_exc())
        finally:
            with contextlib.suppress(Exception):
                self.tick()

        self.root._executing_thread = None
        self.__thread = None
        self.__process = None



#!/usr/bin/env python

from http.client import HTTPConnection

import pytest

from circuits.web import Controller
from circuits.web.client import Client, request

from .helpers import urlopen


class Root(Controller):
    def index(self):
        return 'Hello World!'

    def request_body(self):
        return self.request.body.read()

    def response_body(self):
        return 'ä'

    def request_headers(self):
        return self.request.headers['A']

    def response_headers(self):
        self.response.headers['A'] = 'ä'
        return 'ä'

    def argument(self, arg):
        return arg


def test_index(webapp):
    f = urlopen(webapp.server.http.base)
    s = f.read()
    assert s == b'Hello World!'


@pytest.mark.parametrize(
    'body',
    [
        'ä'.encode(),
        'ä'.encode('iso8859-1'),
    ],
)
def test_request_body(webapp, body):
    connection = HTTPConnection(webapp.server.host, webapp.server.port)
    connection.connect()

    connection.request('POST', '/request_body', body)
    response = connection.getresponse()
    assert response.status == 200
    assert response.reason == 'OK'
    s = response.read()
    assert s == body

    connection.close()


def test_response_body(webapp):
    connection = HTTPConnection(webapp.server.host, webapp.server.port)
    connection.connect()

    connection.request('GET', '/response_body')
    response = connection.getresponse()
    assert response.status == 200
    assert response.reason == 'OK'
    s = response.read()
    assert s == 'ä'.encode()

    connection.close()


def test_request_headers(webapp):
    connection = HTTPConnection(webapp.server.host, webapp.server.port)
    connection.connect()

    body = b''
    headers = {'A': 'ä'}
    connection.request('GET', '/request_headers', body, headers)
    response = connection.getresponse()
    assert response.status == 200
    assert response.reason == 'OK'
    s = response.read()
    assert s == 'ä'.encode()

    connection.close()


def test_response_headers(webapp):
    client = Client()
    client.start()
    client.fire(
        request(
            'GET',
            'http://%s:%s/response_headers'
            % (
                webapp.server.host,
                webapp.server.port,
            ),
        ),
    )

    while client.response is None:
        pass
    assert client.response.status == 200
    assert client.response.reason == 'OK'
    s = client.response.read()
    a = client.response.headers.get('A')
    assert a == 'ä'
    assert s == 'ä'.encode()


def test_argument(webapp):
    connection = HTTPConnection(webapp.server.host, webapp.server.port)
    connection.connect()

    data = 'arg=%E2%86%92'
    connection.request('POST', '/argument', data, {'Content-type': 'application/x-www-form-urlencoded'})
    response = connection.getresponse()
    assert response.status == 200
    assert response.reason == 'OK'
    s = response.read()
    assert s.decode('utf-8') == '\u2192'

    connection.close()



#!/usr/bin/env python
from random import random, seed
from time import sleep, time

import pytest

from circuits.core import Component, Event, Worker, handler, task


class hello(Event):
    """hello Event"""

    success = True


def process(x=None):
    sleep(random())
    return x


class App(Component):
    @handler('hello')
    def _on_hello(self):
        e1 = task(process, 1)
        self.fire(task(process, 2))
        self.fire(task(process, 3))
        yield (yield self.call(e1))


@pytest.fixture()
def app(request, manager, watcher):
    seed(time())

    app = App().register(manager)
    assert watcher.wait('registered')

    worker = Worker().register(manager)
    assert watcher.wait('registered')

    def finalizer():
        app.unregister()
        worker.unregister()

    request.addfinalizer(finalizer)

    return app


def test_call_order(manager, watcher, app):
    x = manager.fire(hello())
    assert watcher.wait('hello_success')

    value = x.value

    assert value == 1



#!/usr/bin/env python
import pytest

from circuits import Component, Event


class hello(Event):
    """hello Event"""

    success = True


class App(Component):
    channel = 'app'

    def hello(self):
        return 'Hello World!'


@pytest.fixture()
def app(request, manager, watcher):
    app = App().register(manager)
    assert watcher.wait('registered')

    def finalizer():
        app.unregister()

    request.addfinalizer(finalizer)

    return app


def test(manager, watcher, app):
    x = manager.fire(hello(), app)
    assert watcher.wait('hello_success')

    value = x.value
    assert value == 'Hello World!'



"""
Controllers

This module implements ...
"""

import json
from collections.abc import Callable
from functools import update_wrapper

from circuits.core import BaseComponent, handler
from circuits.tools import getargspec

from . import tools
from .errors import forbidden, httperror, notfound, redirect
from .wrappers import Response


def expose(*channels, **config):
    def decorate(f):
        @handler(*channels, **config)
        def wrapper(self, event, *args, **kwargs):
            try:
                if not hasattr(self, 'request'):
                    (self.request, self.response), args = args[:2], args[2:]
                    self.request.args = args
                    self.request.kwargs = kwargs
                    self.cookie = self.request.cookie
                    if hasattr(self.request, 'session'):
                        self.session = self.request.session
                if not getattr(f, 'event', False):
                    return f(self, *args, **kwargs)
                return f(self, event, *args, **kwargs)
            finally:
                if hasattr(self, 'request'):
                    del self.request
                    del self.response
                    del self.cookie
                if hasattr(self, 'session'):
                    del self.session

        wrapper.args, wrapper.varargs, wrapper.varkw, wrapper.defaults = getargspec(f)
        if wrapper.args and wrapper.args[0] == 'self':
            del wrapper.args[0]

        if wrapper.args and wrapper.args[0] == 'event':
            f.event = True
            del wrapper.args[0]
        wrapper.event = True

        return update_wrapper(wrapper, f)

    return decorate


class ExposeMetaClass(type):
    def __init__(cls, name, bases, dct):
        super().__init__(name, bases, dct)

        for k, v in dct.items():
            if isinstance(v, Callable) and not (k[0] == '_' or hasattr(v, 'handler')):
                setattr(cls, k, expose(k)(v))


class BaseController(BaseComponent):
    channel = '/'

    @property
    def uri(self):
        """
        Return the current Request URI

        .. seealso:: :py:class:`circuits.web.url.URL`
        """
        if hasattr(self, 'request'):
            return self.request.uri
        return None

    def forbidden(self, description=None):
        """
        Return a 403 (Forbidden) response

        :param description: Message to display
        :type description: str
        """
        return forbidden(self.request, self.response, description=description)

    def notfound(self, description=None):
        """
        Return a 404 (Not Found) response

        :param description: Message to display
        :type description: str
        """
        return notfound(self.request, self.response, description=description)

    def redirect(self, urls, code=None):
        """
        Return a 30x (Redirect) response

        Redirect to another location specified by urls with an optional
        custom response code.

        :param urls: A single URL or list of URLs
        :type urls: str or list

        :param code: HTTP Redirect code
        :type code: int
        """
        return redirect(self.request, self.response, urls, code=code)

    def serve_file(self, path, type=None, disposition=None, name=None):
        return tools.serve_file(
            self.request,
            self.response,
            path,
            type,
            disposition,
            name,
        )

    def serve_download(self, path, name=None):
        return tools.serve_download(
            self.request,
            self.response,
            path,
            name,
        )

    def expires(self, secs=0, force=False):
        tools.expires(self.request, self.response, secs, force)


Controller = ExposeMetaClass('Controller', (BaseController,), {})


def exposeJSON(*channels, **config):
    def decorate(f):
        @handler(*channels, **config)
        def wrapper(self, *args, **kwargs):
            try:
                if not hasattr(self, 'request'):
                    self.request, self.response = args[:2]
                    args = args[2:]
                    self.cookie = self.request.cookie
                    if hasattr(self.request, 'session'):
                        self.session = self.request.session
                self.response.headers['Content-Type'] = 'application/json'
                result = f(self, *args, **kwargs)
                if isinstance(result, (httperror, Response)):
                    return result
                return json.dumps(result)
            finally:
                if hasattr(self, 'request'):
                    del self.request
                    del self.response
                    del self.cookie
                if hasattr(self, 'session'):
                    del self.session

        wrapper.args, wrapper.varargs, wrapper.varkw, wrapper.defaults = getargspec(f)
        if wrapper.args and wrapper.args[0] == 'self':
            del wrapper.args[0]

        return update_wrapper(wrapper, f)

    return decorate


class ExposeJSONMetaClass(type):
    def __init__(cls, name, bases, dct):
        super().__init__(name, bases, dct)

        for k, v in dct.items():
            if isinstance(v, Callable) and not (k[0] == '_' or hasattr(v, 'handler')):
                setattr(cls, k, exposeJSON(k)(v))


JSONController = ExposeJSONMetaClass('JSONController', (BaseController,), {})



#!/usr/bin/env python
from circuits.web import Controller

from .helpers import urlencode, urlopen


class Root(Controller):
    def index(self, *args, **kwargs):
        return f'{args!r}\n{kwargs!r}'


def test(webapp):
    args = ('1', '2', '3')
    kwargs = {'data': '\x00' * 4096}
    url = '%s/%s' % (webapp.server.http.base, '/'.join(args))
    data = urlencode(kwargs).encode('utf-8')
    f = urlopen(url, data)
    data = f.read().split(b'\n')
    assert eval(data[0]) == args
    assert eval(data[1]) == kwargs



#!/usr/bin/env python
"""
Simple Event Filtering

This example shows how you can filter events by using the ``Event.stop()``
method which prevents other event handlers listening to the event from
running. When this example is run you should only get a single line of
output "Hello World!".

.. code-block:: sh

    $ ./filter.py
    Hello World!
"""

from circuits import Component, Event


class hello(Event):
    """hello Event"""


class App(Component):
    def hello(self, event):
        """Hello Event Handler"""
        event.stop()  # Stop further event processing

        print('Hello World!')

    def started(self, event, component):
        """
        Started Event Handler

        This is fired internally when your application starts up
        and can be used to trigger events that only occur once
        during startup.
        """
        event.stop()  # Stop further event processing
        self.fire(hello())  # Fire a Hello event
        raise SystemExit(0)  # Terminate the application


# Start and "run" the system.
# We're deliberately creating two instances of ``App``
# so we can demonstrate event filtering.
app = App()
App().register(app)  # 2nd App
app.run()



"""
Utilities

This module implements utility functions.
"""

import os
import re
import stat
import struct
import time
import zlib
from math import sqrt
from urllib.parse import parse_qs as _parse_qs

from .exceptions import RangeUnsatisfiable


image_map_pattern = re.compile('^[0-9]+,[0-9]+$')


def is_unix_socket(path):
    if not os.path.exists(path):
        return False

    mode = os.stat(path).st_mode

    return stat.S_ISSOCK(mode)


def average(xs):
    return sum(xs) * 1.0 / len(xs)


def variance(xs):
    avg = average(xs)
    return [(x - avg) ** 2 for x in xs]


def stddev(xs):
    return sqrt(average(variance(xs)))


def parse_qs(query_string, keep_blank_values=True):
    """
    parse_qs(query_string) -> dict

    Build a params dictionary from a query_string.
    If keep_blank_values is True (the default), keep
    values that are blank.
    """
    if image_map_pattern.match(query_string):
        # Server-side image map. Map the coords to "x" and "y"
        # (like CGI::Request does).
        pm = query_string.split(',')
        return {'x': int(pm[0]), 'y': int(pm[1])}
    pm = _parse_qs(query_string, keep_blank_values)
    return {k: v[0] for k, v in pm.items() if v}


def compress(body, compress_level):
    """Compress 'body' at the given compress_level."""
    # Header
    yield b'\037\213\010\0' + struct.pack('<L', int(time.time())) + b'\002\377'

    size = 0
    crc = zlib.crc32(b'')

    zobj = zlib.compressobj(
        compress_level,
        zlib.DEFLATED,
        -zlib.MAX_WBITS,
        zlib.DEF_MEM_LEVEL,
        0,
    )

    for chunk in body:
        if not isinstance(chunk, bytes):
            chunk = chunk.encode('utf-8')

        size += len(chunk)
        crc = zlib.crc32(chunk, crc)
        yield zobj.compress(chunk)

    yield zobj.flush() + struct.pack('<l', crc) + struct.pack('<L', size & 0xFFFFFFFF)


def get_ranges(headervalue, content_length):
    """
    Return a list of (start, stop) indices from a Range header, or None.

    Each (start, stop) tuple will be composed of two ints, which are suitable
    for use in a slicing operation. That is, the header "Range: bytes=3-6",
    if applied against a Python string, is requesting resource[3:7]. This
    function will return the list [(3, 7)].

    If this function returns an empty list, you should return HTTP 416.
    """
    if not headervalue:
        return None

    result = []
    _bytesunit, byteranges = headervalue.split('=', 1)
    for brange in byteranges.split(','):
        start, stop = (x.strip() for x in brange.split('-', 1))
        if start:
            if not stop:
                stop = content_length - 1
            start, stop = list(map(int, (start, stop)))
            if start >= content_length:
                # From rfc 2616 sec 14.16:
                # "If the server receives a request (other than one
                # including an If-Range request-header field) with an
                # unsatisfiable Range request-header field (that is,
                # all of whose byte-range-spec values have a first-byte-pos
                # value greater than the current length of the selected
                # resource), it SHOULD return a response code of 416
                # (Requested range not satisfiable)."
                continue
            if stop < start:
                # From rfc 2616 sec 14.16:
                # "If the server ignores a byte-range-spec because it
                # is syntactically invalid, the server SHOULD treat
                # the request as if the invalid Range header field
                # did not exist. (Normally, this means return a 200
                # response containing the full entity)."
                return None
            # Prevent duplicate ranges. See Issue #59
            if (start, stop + 1) not in result:
                result.append((start, stop + 1))
        else:
            if not stop:
                # See rfc quote above.
                return None
            # Negative subscript (last N bytes)
            # Prevent duplicate ranges. See Issue #59
            if (content_length - int(stop), content_length) not in result:
                result.append((content_length - int(stop), content_length))

    # Can we satisfy the requested Range?
    # If we have an exceedingly high standard deviation
    # of Range(s) we reject the request.
    # See Issue #59

    if len(result) > 1 and stddev([x[1] - x[0] for x in result]) > 2.0:
        raise RangeUnsatisfiable()

    return result



from circuits import BaseComponent, handler
from circuits.net.events import close, connect
from circuits.net.sockets import TCPClient

from .protocol import Protocol


class Client(BaseComponent):
    """Node Client (peer)"""

    channel = 'node_client'

    def __init__(self, host, port, channel=channel, receive_event_firewall=None, send_event_firewall=None, **kwargs):
        """
        Create new connection for a node.

        :param hostname:    hostname to connect.
        :type hostname:     str

        :param port:    port to connect.
        :type port:     int

        :param channel: An optional keyword argument which if defined,
                        set channel used for node event.
                        **Default:** ``node_client``
        :type channel:  str

        :param receive_event_firewall:  An optional keyword argument which if
                                        defined, function or method to call for
                                        check if event is allowed for sending.
                                        **Default:** ``None`` (no firewall)
        :type receive_event_firewall:   function
        :type receive_event_firewall:   method

        :param send_event_firewall:  An optional keyword argument which if
                                    defined, function or method to call for
                                    check if event is allowed for executing
                                    **Default:** ``None`` (no firewall)
        :type send_event_firewall:   function
        :type send_event_firewall:   method
        """
        super().__init__(channel=channel, **kwargs)

        self.__host = host
        self.__port = port
        self.__protocol = Protocol(
            receive_event_firewall=receive_event_firewall,
            send_event_firewall=send_event_firewall,
            channel=channel,
        ).register(self)

        TCPClient(channel=channel, **kwargs).register(self)

    @handler('ready')
    def _on_ready(self, component):
        self.connect()

    def close(self):
        """Close the connection"""
        self.fire(close())

    def connect(self):
        """Create the connection"""
        self.fire(connect(self.__host, self.__port))

    def send(self, event):
        """
        Send event through the connection

        :param event:    Event to send.
        :type event:     :class:`circuits.core.events.Event`

        :return: The result object of the sended event
        :rtype: generator
        """
        return self.__protocol.send(event)

    @handler('read')
    def _on_read(self, data):
        self.__protocol.add_buffer(data)
